/* A general model combining Tuffley-Steel, Huelsenbeck and Galtier models */
/* This code was based on compute_ml.c in NHML (Galtier & Gouy, 1998; Galtier, 2001)  */

/* chunks of dead code are effectively commented out with
 * ifdef COMPILE_DEAD_CODE
 * hopefully this will facilitate removing struct members that aren't used, etc.
 */

#include "mynhmlg.h"
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>

extern char *siteLKHfileName;
//double pi, covar1, covar2, covar3; /* global variables */

char aachar[]="ARNDCQEGHILKMFPSTWYVBZX?*-";

static const double min_value[4]={frmin, lmin, covmin, pimin};
static const double max_value[4]={frmax, lmax, covmax, pimax};

static const double gamma_range[GAMMA_RANGE_SIZE]={0.1, 0.2, 0.4, 0.6, 0.8, 1., 2., 5., 10.};
double *gamma_clmean; // discrete gamma lookup table (by class)
static double gamma_shp; // was a local var in compute()



/***************** initialize branch lengths ****************/
/* initialize branch lengths from a pairwise distance matrix.*/

/* matvect */
/* Return the product vector of matrix m1(nbl1, nbc1) by column vector v2. */
static int matvect(int nbl1, int nbc1, double m1[nbl1][nbc1], double *v2, int nbl2, double *vprod)
{
  int i, k;
  if(nbc1!=nbl2) return 0;

  for(i=0;i<nbl1;i++){
    vprod[i]=0;
    for(k=0;k<nbc1;k++) vprod[i]+=m1[i][k]*v2[k];
  }
  return 1;
}

/* vectmat */
/* Return the product vector of row vector v1 by matrix m2(nbl1, nbc1). */
static double* vectmat(double* v1, int nbc1, int nbl2, int nbc2, double m2[nbl2][nbc2])
{
  int j, k;
  if(nbc1!=nbl2) return 0;
  double *vprod = check_alloc(nbc2+1, sizeof(double));
  for(j=0;j<nbc2;j++){
    vprod[j]=0;
    for(k=0;k<nbc1;k++) vprod[j]+=v1[k]*m2[k][j];
  }
  return vprod;
}



/* lslgbr */
/* Compute the least-square estimates of branch lengths of tree btree */
/* according to effective distances dist. Branch length are returned */
/* via a double array, ordered as in btree. sce returned value is the sum */
/* of squares of residuals between patristic and effective distances. */
/* totallength returned value is the sum of all branch lengths. */
/* Negative branch lengths are forced to 0 in sce and tot computing. */

static double* lslgbr(int** btree, int n, double** dist, double* sce, double* totallength)
{
  int i, j, k, b1, b2;
  int nbdist=n*(n-1)/2, nbbr=2*n-3;
  double ta[nbbr][nbdist], B[nbbr][nbbr]; // matrices on the stack
  double *lgbr, *pdist; // vectors

  double *invB  = check_alloc(nbbr*nbbr, sizeof(*invB));
  double *prod1 = check_alloc(nbbr*nbdist, sizeof(*prod1));
  double *cdist = check_alloc(nbdist, sizeof(double));
  int *ind1 = check_alloc(nbdist, sizeof(*ind1));
  int *ind2 = check_alloc(nbdist, sizeof(*ind2));

  int ii=0;	// cdist = vec(dist), i.e. concat into a single vector
  for(i=0;i<n-1;i++){
    for(j=i+1;j<n;j++){
      cdist[ii]=dist[i][j];
      ind1[ii]=i; ind2[ii]=j;
      ii++;
    }
  }

  /* build ta */
  for(i=0;i<n;i++) {
    for(j=0;j<nbdist;j++){
      if(ind1[j]==i || ind2[j]==i) ta[i][j]=1.; else ta[i][j]=0.;
    }
  }
  for(i=n;i<nbbr;i++){
    for(j=0;j<nbdist;j++){
      b1=testbit(btree[i-n],ind1[j]+1); b2=testbit(btree[i-n],ind2[j]+1);
      if((b1 && !b2) || (b2 && !b1)) ta[i][j]=1.; else ta[i][j]=0.;
    }
  }

  /* compute B */
  for(i=0;i<nbbr;i++) {
    for(j=0;j<nbbr;j++){
      B[i][j]=0.;
      for(k=0;k<nbdist;k++) B[i][j]+=ta[i][k]*ta[j][k];
    }
  }

  /* compute lgbr and pdist */
  lgbr=(double*)check_alloc(nbbr, sizeof(double));
  if(!invmat(nbbr, (double *)B, invB)) return NULL;
  matmat(nbbr, nbbr, invB, nbbr, nbdist, (double *)ta, prod1);
  matvect(nbbr, nbdist, (double (*)[nbdist])prod1, cdist, nbdist, lgbr);
  pdist=vectmat(lgbr, nbbr, nbbr, nbdist, ta);

  /* compute sce and total length */

  if(sce){
    *sce=0.;
    for(i=0;i<nbdist;i++)
      if(pdist[i]>0.) *sce+=(pdist[i]-cdist[i])*(pdist[i]-cdist[i]);
      else *sce+=cdist[i]*cdist[i];
  }

  if(totallength){
    *totallength=0.;
    for(i=0;i<nbbr;i++) if(lgbr[i]>0.) *totallength+=lgbr[i];
  }

  free(cdist); free(ind1); free(ind2);
  free(invB); free(prod1);
  free(pdist);
  return lgbr;
}


static int congruent(noeud nd, noeud from, int** btree, char** name, int nbseq, int j)
{
  noeud to1, to2;
  int i;

  if(nd==root){
    if(from==root->v1) return congruent(root->v2, root, btree, name, nbseq, j);
    else if(from==root->v2) return congruent(root->v1, root, btree, name, nbseq, j);
  }

  if(nd->v1==NULL){
    for(i=0;i<nbseq;i++){
      if(samename(nd->nom, name[i])){
	if(testbit(btree[j], i+1)) return 1; else return 0;
      }
    }
    if(i==nbseq){ fprintf(stderr, "procov: cong error at: %s\n", __func__); exit(3); }
  }

  if(from==nd->v1) {to1=nd->v2; to2=nd->v3;}
  else if(from==nd->v2) {to1=nd->v1; to2=nd->v3;}
  else if(from==nd->v3) {to1=nd->v1; to2=nd->v2;}
  else if(from==root->v1 && nd==root->v2) {to1=nd->v1; to2=nd->v2;}
  else if(from==root->v2 && nd==root->v1) {to1=nd->v1; to2=nd->v2;}
  else { fprintf(stderr, "procov: cong2 error: nd=%s , from=%s\n", nd->nom, from->nom); exit(3);}

  int c1 = congruent(to1, nd, btree, name, nbseq, j);
  int c2 = congruent(to2, nd, btree, name, nbseq, j);

  if(c1==c2) return c1;
  return 10;
}


static void btltostl(tree* stree, int** btree, char** name, double* treelg){

  int i, j, left, right;
  int nbseq=stree->nbseq;
  branche br;
  noeud boutterm;

  for(i=0;i<2*nbseq-3;i++){
    br=stree->listbr[i];
    if(br->bout1->v1==NULL || br->bout2->v1==NULL){ /* terminal */
      boutterm=(br->bout1->v1)?br->bout2:br->bout1;
      for(j=0;j<nbseq;j++){
	if(samename(boutterm->nom, name[j])){
	  br->length=treelg[j];
    	  break;
	}
      }
      if(j==nbseq) fprintf(stderr, "procov: error1 at: %s\n", __func__);
    }
    else{					/* internal */
      for(j=0;j<nbseq-3;j++){
        left=congruent(br->bout1, br->bout2, btree, name, nbseq, j);
        right=congruent(br->bout2, br->bout1, btree, name, nbseq, j);
	if((left==0 && right==1) || (left==1 && right==0)){
	  br->length=treelg[nbseq+j];
	  break;
	}
      }
      if(j==nbseq-3) fprintf(stderr, "procov: error2 at: %s\n", __func__);
    }
  }
  reset_bl(stree, root);
}



/* unroot_c */
/* Remove root from rooted c_tree carbre. */
static int unroot_c(char* carbre)
{
  int diff, i=0;
  while((carbre[i]==' ' || carbre[i]=='\n' || carbre[i]=='\t') && carbre[i]) i++;
  if(carbre[i]!='[' && carbre[i]!='(') {
    fputs("procov: error: Tree first char must be ( or [\n", stderr);
    return -1;
  }

  if(carbre[i]=='[') while(carbre[i]!=']' && carbre[i]) i++;
  if (!carbre[i]){ fputs("procov: error: Unmatched '[' ']'\n", stderr); return -1; }
  while(carbre[i]!='(' && carbre[i]) i++;
  if (!carbre[i]){ fputs("No initial parenthesis\n", stderr); return -1; }

  i++;
  while(carbre[i]!='(' && carbre[i] && carbre[i]!=';') i++;
  if(!carbre[i] || carbre[i]==';') return 0;
  carbre[i]=' ';
  diff=0;
  while(diff!=-1 && carbre[i] && carbre[i]!=';'){
    if(carbre[i]=='(')  diff++;
    if(carbre[i]==')')  diff--;
    i++;
  }

  if(!carbre[i] || carbre[i]==';') return 0;
  carbre[i-1]=' ';
  while(carbre[i]!=',' && carbre[i]!=')') {
    carbre[i]=' ';
    i++;
  }

  return 1;
}



void init_bl(tree* stree)
{
  int i, nbseq = stree->nbseq, lliste = (nbseq+lmot-1)/lmot;

  char *ctree = check_alloc(50*nbseq, sizeof(char));
  int **btree = check_alloc(nbseq-3, sizeof(int*));
  for(i=0;i<nbseq-3;i++)
    btree[i]=(int*)check_alloc(lliste, sizeof(int));
  double *newbl=(double*)check_alloc(2*nbseq-3, sizeof(double));

  stoc(stree->node, TRUE, 1, ctree);
  unroot_c(ctree);
  ctob(ctree, stree->names, btree);

  newbl=lslgbr(btree, nbseq, stree->dist, NULL, NULL);
  for(i=0;i<2*nbseq-3;i++){
    if(newbl[i]<lmin) newbl[i]=lmin;
    if(newbl[i]>lmax) newbl[i]=lmax;
  }

  double sumalive=0, sum=0;
  for(i=0;i<2*nbseq-3;i++){
    sum+=newbl[i];
    if(stree->alivebr[i]) sumalive+=newbl[i];
  }
  double rap = sum/sumalive;
  for(i=0;i<2*nbseq-3;i++)
    if(stree->alivebr[i])
      newbl[i]*=rap;


  btltostl(stree, btree, stree->names, newbl);

  for(i=0;i<nbseq-3;i++) free(btree[i]);
  free(btree);
  free(ctree);
  free(newbl);
}


int setunderlyingbr(noeud nd, branche br, int i){

  int value=0;

  if(nd->v1==NULL){
    nd->underlyingbr[i]=0;
    return 0;
  }

  if(setunderlyingbr(nd->v1, br, i)) value=1;
  else if(setunderlyingbr(nd->v2, br, i)) value=2;
  else if(nd==br->bout1 && nd->v1==br->bout2) value=10;
  else if(nd==br->bout2 && nd->v1==br->bout1) value=10;
  else if(nd==br->bout1 && nd->v2==br->bout2) value=20;
  else if(nd==br->bout2 && nd->v2==br->bout1) value=20;
  else if(nd->v1==br->bout1 && nd->v2==br->bout2) value=12;  /* nd=root, br=root branch */
  else if(nd->v2==br->bout1 && nd->v1==br->bout2) value=12;  /* nd=root, br=root branch */
  
  nd->underlyingbr[i]=value;

  return value;
}


static void fprinttime(double rtime, FILE *out)
{
  int d, h, m, s;

  #define plural(t) (t>1)?"s":""
  d=(int)rtime/86400;
  rtime-=d*86400;
  if(d) fprintf(out,"%d day%s ", d, plural(d));

  h=(int)rtime/3600;
  rtime-=h*3600;
  if(h) fprintf(out,"%d hour%s ", h, plural(h));

  m=(int)rtime/60;
  rtime-=m*60;
  if(m) fprintf(out,"%d minute%s ", m, plural(m));

  s=(int)rtime;
  if(s) fprintf(out,"%d second%s ", s, plural(s));

  if(d==0 && h==0 && m==0 && s==0)
    fprintf(out,"less than 1 second");

  fputc('\n', out);
}


void bit1(int *plist, int num){
  num--;
  plist+=(num/lmot);
  *plist |= (1<<(num%lmot));
}


static void et(int *listet, int *list1, int *list2, int len){
  int i;
  for(i=0; i< len; i++)
	listet[i]= list1[i] & list2[i];
}
static void ou(int *listou, int *list1, int *list2, int len){
  int i;
  for(i=0; i< len; i++)
    listou[i]= list1[i] | list2[i];
}
static void non(int *listnon, int *list, int len){
  int i;
  for(i=0; i< len; i++)
	listnon[i]= ~list[i];
}

static FILE* dbg, *dbgprov;
static void printmemory_init(){
  char *line, *ret;
  int maxlline=1000;


  dbg=fopen("debugNHML", "w");
  system("ps -aux > debugNHML.prov");
  dbgprov=fopen("debugNHML.prov", "r");
  if(dbgprov==NULL){printf("Warning: problem when debugging\n"); return;}
  line=check_alloc(maxlline+1, sizeof(char));
  ret=fgets(line, maxlline, dbgprov);
  if(ret!=NULL) fprintf(dbg, "             %s", line);

  fclose(dbgprov);
  fclose(dbg);
  free(line);
}

static void printmemory(char title_arg[]){

  char* line, *ret, *syscall, *title, *prov;
  int maxlline=1000;
  int thispid, lgtit=10, lgarg, i;

  thispid=(int)getpid();
  lgarg=(int)strlen(title_arg);
  title=check_alloc(lgtit+3, sizeof(char));
  sprintf(title, "%.10s", title_arg);
  prov=title;
  while(*prov) prov++;
  for(i=0;i<lgtit-lgarg;i++){*prov=' '; prov++;}


  syscall=check_alloc(500, sizeof(char));
  sprintf(syscall, "ps -aux | grep eval_nhg | grep %d > debugNHML.prov", thispid);

  system(syscall);

  dbgprov=fopen("debugNHML.prov", "r");
  dbg=fopen("debugNHML", "a");
  line=check_alloc(maxlline+1, sizeof(char));

  while(1){
    ret=fgets(line, maxlline, dbgprov);
    if(ret==NULL) break;
    fprintf(dbg, "%s: %s", title, line);
  }

  fclose(dbgprov);
  fclose(dbg);
  system("rm -f debugNHML.prov");
  free(line);
  free(syscall);
  free(title);
}


/* matmat: Product of matrices. */
/* prod(nbl1, nbc2) = m1(nbl1, nbc1) * m2(nbl2, nbc2) */
#ifdef USE_BLAS
/* This is significantly faster, esp. with larger arrays (gamma_nbcl>1) */
inline int matmat(int nbl1, int nbc1, double *m1, int nbl2, int nbc2, double *m2, double *prod)
{
  if(nbc1!=nbl2)  return 0;
#ifdef HAVE_CBLAS
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, nbl1, nbc2, nbl2,
	      1.0, m1, nbc1, m2, nbc2, 0.0, prod, nbc2 );
#else
  // C = AB  is equivalent to  tr(C) = tr(B)*tr(A)
  dgemm(				       'N', 'N', nbc2, nbl1, nbl2,
	      1.0, m2, nbc2, m1, nbc1, 0.0, prod, nbc2 );
#endif
  return 1;
}
#else
/* Without BLAS, this function uses a large fraction (~70%) of the
 * total run time of procov on the sample cox2 data set.  It is called 22144 times,
 * counting the inlined copies in rtop. array dim = 40x40. */
int matmat(int nbl1, int nbc1, double *M1, int nbl2, int nbc2, double *M2, double *PROD)
{
  int i, j, k;
  // yes, it's possible to avoid using pos() with clever hacking.  We do this instead of
  // having var len array function args to keep the signature matching blas_matmat
  // cdecl was very useful for figuring out how to write the declarations and casts...
  double (*m1)[nbc1], (*m2)[nbc2], (*prod)[nbc2];
  m1 = (double (*)[])M1, m2 = (double (*)[])M2, prod = (double (*)[])PROD;
  if(nbc1!=nbl2) return 0;

  for(i=0;i<nbl1;i++) for(j=0;j<nbc2;j++){
    double p = 0;
    for(k=0;k<nbc1;k++)
      p += m1[i][k] * m2[k][j];
    prod[i][j] = p;
  }
  return 1;
}
#endif // USE_BLAS

#ifdef USE_LAPACK
int invmat(int n, double *mat, double *invmat)
{
  int ipiv[n];
#ifdef HAVE_CLAPACK
  memcpy(invmat, mat, n*n*sizeof(double));
  if     (clapack_dgetrf(CblasRowMajor, n,n, invmat, n, ipiv)) return 0; // LU
  return !clapack_dgetri(CblasRowMajor, n,   invmat, n, ipiv); // LU->inverse
#else
  // Use the full-blown Fortran interface: Intel's non _ dgetrf still needs a work array
  int i, j, ret, lwork=80*n; // optimal is n*optimal block size
#if LAPACK_MALLOC
  double *colmajor = check_alloc(n*n,   sizeof(double));
  double *work     = check_alloc(lwork, sizeof(double));
#else
  double colmajor[n*n];
  double work[lwork];
#endif
  for(i=0 ; i<n ; i++) for(j=0 ; j<n ; j++)
    colmajor[pos(j,i,n)] = mat[pos(i,j,n)];
  dgetrf_(&n,&n, colmajor, &n, ipiv, &ret); // LU
  if(ret) return 0;
  dgetri_(&n,   colmajor, &n, ipiv, work, &lwork, &ret); // LU->inverse
  for(i=0 ; i<n ; i++) for(j=0 ; j<n ; j++)
    invmat[pos(i,j,n)] = colmajor[pos(j,i,n)];
#ifdef LAPACK_MALLOC
  free(colmajor); free(work);  // leaks on early return
#endif
  return !ret;
#endif
}

#else
/* ludcmp */
/* from Numerical Recipes in C */
/* Replace matrix a by a rowwise permutation of its LU decomposition. */

static int ludcmp(double **a, int n, int *indx, double *d){
	int i,imax,j,k;
	double big,dum,sum,temp;
	double *vv;

	imax=-999; // keep gcc happy
	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) return 0;
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=1.0e-20;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free_vector(vv,1,n);
	return 1;
}



/* lubksb */
/* from Numerical Recipes in C */

static void lubksb(double **a, int n, int *indx, double b[]){
	int i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

/* invmat */
/* Invert square matrix mat(n,n). Result in invmat.*/
/* If mat is singular, 0 is returned (1 otherwise) */
int invmat(int n, double *arg_mat, double *arg_invmat)
{
  int i, j, *indx;
  double **lu, d, *col;
  double (*mat)[n] = (double (*)[])arg_mat, (*invmat)[n] = (double (*)[])arg_invmat;

  lu=check_alloc(n+1, sizeof(double*));
  for(i=1;i<n+1;i++){
    lu[i]=(double*)check_alloc(n+1, sizeof(double));
    for(j=1;j<n+1;j++) lu[i][j]=mat[i-1][j-1];
  }
  indx=(int*)check_alloc(n+1, sizeof(int));
  col=(double*)check_alloc(n+1, sizeof(double));

  if(!ludcmp(lu, n, indx, &d)) return 0;

  for(j=1;j<=n;j++){
    for(i=1;i<=n;i++) col[i]=0.;
    col[j]=1.;
    lubksb(lu, n, indx, col);
    for(i=1;i<=n;i++) invmat[i-1][j-1]=col[i];
  }
  for(i=1; i<n+1;i++) free(lu[i]);
  free(lu);
  free(indx); free(col);
  return 1;
}
#endif //USE_LAPACK

/* Diagonalize square matrix A (dim=n), which is destroyed. */
/* eigen values returned in rr (real part only) */
/* right eigen vectors returned as vr (real part only) */

static int callMyEigen(int n, double *A, double *vr, double *rr)
{
  double w[2*n], ri[n], vi[n*n]; // imaginary parts: discard
  int ret=eigen(1, A, n, rr, ri, vr, vi, w);
  if (ret==-1) fputs("procov: Diagonalization problem: not converged\n", stderr);
  if (ret==1)  fputs("procov: error: complex eigen values\n", stderr);
  return ret;
}




/* findroot */
/* Return the root node of rooted s_tree including node nd. */
/* or NULL if no root exists */
noeud findroot(noeud p, int taxa){
  int i;
  for( i=2*taxa ; i>0  ; i-- ){ // inf loop countermeasures
    if (p->v3) p=p->v3;
    else break;
  }
  if(!i) return NULL;
  return p;
}

#ifdef COMPILE_DEAD_CODE
/* mindepth */
/* Set node's depth in a rooted tree. */
/* mindepth(terminal node) = 0 */
/* mindepth(internal node) = number of branches between node and closest terminal node */
/* node comesfrom (one of node's neighbors) is root. */

static int mindepth (noeud comesfrom, noeud node){
  noeud droit, gauche;

  if (node->v1==NULL) { node->depth=0; return 0; }
  if (node->v2==NULL) { node->depth=0; return 0; }

  if (comesfrom==node->v1) {
    gauche=node->v2;
    droit=node->v3;
  }
  else if (comesfrom==node->v2) {
    gauche=node->v1;
    droit=node->v3;
  }
  else {
    gauche=node->v1;
    droit=node->v2;
  }

  node->depth=mini(mindepth(node, gauche), mindepth(node, droit))+1;
  return node->depth;
}
#endif // COMPILE_DEAD_CODE

static void invert_branch(branche br){
  noeud prov;
  prov=br->bout1; br->bout1=br->bout2; br->bout2=prov;
}

static void organize_listbr_node(noeud nd, branche* listbr, int nbseq)
{
  int i, break1=0, break2=0;
  if(nd->v1==NULL) return;

  for(i=0;i<nbseq;i++){
    if(!break1 && (listbr[i]->bout1==nd->v1 && listbr[i]->bout2==nd)){
      invert_branch(listbr[i]);
      break1=1;
    }
    if(!break1 && (listbr[i]->bout1==nd && listbr[i]->bout2==nd->v1)){
      break1=1;
    }
    if(!break2 && (listbr[i]->bout1==nd->v2 && listbr[i]->bout2==nd)){
      invert_branch(listbr[i]);
      break2=1;
    }
    if(!break2 && (listbr[i]->bout1==nd && listbr[i]->bout2==nd->v2)){
      break2=1;
    }
    if(break1 && break2) break;
  }

  organize_listbr_node(nd->v1, listbr, nbseq);
  organize_listbr_node(nd->v2, listbr, nbseq);
}

void organize_listbr(tree* s_tree)
{
  int i, nbseq=s_tree->nbseq;
  branche* listbr=s_tree->listbr;

  for(i=0;i<2*nbseq-3;i++){
    if(listbr[i]->bout1==root->v2 && listbr[i]->bout2==root->v1){
      invert_branch(listbr[i]);
      break;
    }
  }
  organize_listbr_node(root->v1, listbr, nbseq);
  organize_listbr_node(root->v2, listbr, nbseq);
}



static int setpattern_site(noeud nd, int site)
{
  int i, son1, son2;
  if(nd->v1==NULL){
    // maybe strchr(c, aachar);
    if (nd->seq[site]=='A') return 0;
    else if (nd->seq[site]=='R') return 1; else if (nd->seq[site]=='N') return 2;
    else if (nd->seq[site]=='D') return 3; else if (nd->seq[site]=='C') return 4;
    else if (nd->seq[site]=='Q') return 5; else if (nd->seq[site]=='E') return 6;
    else if (nd->seq[site]=='G') return 7; else if (nd->seq[site]=='H') return 8;
    else if (nd->seq[site]=='I') return 9; else if (nd->seq[site]=='L') return 10;
    else if (nd->seq[site]=='K') return 11; else if (nd->seq[site]=='M') return 12;
    else if (nd->seq[site]=='F') return 13; else if (nd->seq[site]=='P') return 14;
    else if (nd->seq[site]=='S') return 15; else if (nd->seq[site]=='T') return 16;
    else if (nd->seq[site]=='W') return 17; else if (nd->seq[site]=='Y') return 18;
    else if (nd->seq[site]=='V') return 19; else if (nd->seq[site]=='B') return 20;
    else if (nd->seq[site]=='Z') return 21;
    return 22;    // this is for the cases of X?*-
  }

  son1=setpattern_site(nd->v1, site);
  son2=setpattern_site(nd->v2, site);

  for(i=0;i<nd->nbpattern;i++)
    if(nd->patternson1[i]==son1 && nd->patternson2[i]==son2)
      return i;

  nd->patternson1[nd->nbpattern]=son1;
  nd->patternson2[nd->nbpattern]=son2;
  nd->nbpattern++;
  return nd->nbpattern-1;
}


void setpattern(tree* s_tree){
  int i, j, n, nbnd;
  static int nbappel=0;

  nbappel++;

  nbnd=2*s_tree->nbseq-2;
  for(i=0;i<nbnd;i++)
    s_tree->node[i]->nbpattern=0;
  root->nbpattern=0;

  for(i=0;i<s_tree->lgseq;i++)
    setpattern_site(root, i);

  assert (s_tree->lgseq == root->nbpattern); // CovarionSiteLikelihood loops over lgseq patterns in the root

  n = 23;		  // pattern values are [0..22] for leaf nodes
  like *leafblock = check_alloc(gamma_nbcl1*n, sizeof(like));
  like **leafx = check_alloc(n, sizeof(like*));
  for(j=0 ; j<n ; j++) // pointers to columns, for the code that's expecting an array-of-pointers
    leafx[j] = &leafblock[j*gamma_nbcl1];

  for(i=0 ; i<=nbnd ; i++){ // for each node, allocate an ->x big enough to hold all its patterns
    noeud nd = (i<nbnd)? s_tree->node[i] : root;
    if(i<s_tree->nbseq){
      nd->x = leafx; // leaf nodes all have their x set the same, so why not share
      continue; 		// don't need ps1size either
    }

    n = nd->nbpattern;
    like *p = check_alloc(gamma_nbcl1*n, sizeof(like));
    nd->x = check_alloc(n, sizeof(like*));
    for(j=0 ; j<n ; j++)	// pointers to columns, for the code that's expecting an array-of-pointers
      nd->x[j] = &p[j*gamma_nbcl1];

    // set up nd->ps1/2size.
    nd->ps1size = nd->ps2size = 0;
    for(j=0 ; j<nd->nbpattern ; j++){
      nd->ps1size = max(nd->ps1size, nd->patternson1[j]);
      nd->ps2size = max(nd->ps2size, nd->patternson2[j]);
    }
    nd->ps1size++; nd->ps2size++; // we want sizes, not indices into 0-based arrays...
    if(debug>=2 && nd->v1->nbpattern > 0 && nd->ps1size != nd->v1->nbpattern)
      fprintf(stderr, "node %s: ps1size: %d, v1->nbpattern: %d\n", nd->nom, nd->ps1size, nd->v1->nbpattern);
    if(debug>=2 && nd->v2->nbpattern > 0 && nd->ps2size != nd->v2->nbpattern)
      fprintf(stderr, "node %s: ps2size: %d, v2->nbpattern: %d\n", nd->nom, nd->ps2size, nd->v2->nbpattern);
  }
}




// set up a mapping from nodes to the table of branches which is used reset_bl
void set_numerobr(noeud nd, branche* listbr, int nbbr)
{
  if(nd==root)
    nd->numerobr=-1;
  else{
    int i;
    for(i=0;i<nbbr;i++){
      noeud br1 = listbr[i]->bout1;
      noeud br2 = listbr[i]->bout2;
      if(nd->v3==root){
	if((root->v1==br1 && root->v2==br2)||(root->v2==br1 && root->v1==br2)){
	  nd->numerobr=i;
	  break;
	}
      }
      else if((nd->v3==br1 && nd==br2)||(nd==br1 && nd->v3==br2)){
	nd->numerobr=i;
	break;
      }
    }
  }

  if(!nd->v1) return;
  set_numerobr(nd->v1, listbr, nbbr);
  set_numerobr(nd->v2, listbr, nbbr);
}

#ifdef COMPILE_DEAD_CODE
static void set_parambl(tree* s_tree, noeud nd){

  double l;

  if(nd==NULL) return;


  if(nd!=root){
    if(nd->v3==root) l=root->l1+root->l2;
    else l=nd->l3;
    s_tree->listbr[nd->numerobr]->length=l;

  }

  set_parambl(s_tree, nd->v1);
  set_parambl(s_tree, nd->v2);
}
#endif // COMPILE_DEAD_CODE

/* Returns rij, ie the ratio of effective length to length ?? */
/* for a branch of length lg, covar rate nu, initial and final */
/* rate classes cl1 and cl2 */
/* and derivatievs with respect to x=lg.nu */
/* my note: rij (w.r.t. lg and nu) is the conitional average rate, see equation 10 of Galtier paper */

static void effective_length(double lg, int cl1, int cl2, double nu, double *rij, double *drij, double* d2rij)
{
  double A, B, N;
  int g;
  if(cl1==cl2) g=gamma_nbcl;
  else g=0;
  double ri=gamma_clmean[cl1];
  double rj=gamma_clmean[cl2];
  double r = ri+rj;
  double a = 2-r;
  double b = 1+(g-2)*r/2;

  double x = lg*nu, emx=exp(-x);
  if(x<=0){                       // nu <=0 means there is no covarion
    *rij=r/2.;                    // because ri=rj for no covarion, rij = ri = rj
    *drij=*d2rij=0.;
    return;
  }
  double em2x=emx*emx, em3x=em2x*emx, em4x=em2x*em2x;
  double x2 = x*x,     x3 = x2*x,     x4 = x2*x2;
  double D = x*(1+(g-1)*emx);
	/* rij */
  if(x<=0.05){   /* devlp limite */
    if(cl1!=cl2)
      *rij=r/2. + (2.-r)*x/12. - (2.-r)*x*x*x/720.;
    else
      *rij=r/2. + (r/2.-1.)*(3.*g-2.)*x*x/(6.*g);

  }
  else{  /* exact calculation */
    N=(a+b*x)*emx+x-a;
    *rij=N/D;
  }

  if(*rij<0.) *rij=0.000001;


	/* drij/dx */
  if(x<=0.05){   /* devlp limite */
    if(cl1!=cl2)
      *drij=(2.-r)/12. - (2.-r)*x*x/240.;
    else
      *drij=(1.-r/2.)*x/(3.*g);
  }
  else{  /* exact calculation */
    A = -a*(g-1);
    B = (g-1-b)*x2 - a*g*x + a*(g-2);
    N = A*em2x + B*emx + a;
    *drij = N/(D*D);
  }


	/* d2rij/dx2 */
  if(x<=0.05){   /* devlp limite */
    if(g==0)
      *d2rij=(r-2.)*x/120.;
    else
      *d2rij=(1.-r/2.)/(3.*g);
  }
  else{  /* exact calculation */
    double gm1 = g-1, gm12=gm1*gm1, gm13=gm12*gm1;
    double P4 = 2*a*(gm13)*x;
    double P3 = gm12*(gm1-b)*x4 - g*gm12*a*x3 + 2.*g*gm12*a*x2 - 2.*(g-4.)*gm12*a*x;
    double P2 = 4*g*gm1*a*x2 - 6*(g-2)*gm1*a*x;
    double P1 = (b-gm1)*x4+g*a*x3+2.*g*a*x2-2.*(3.*g-4.)*a*x;
    *d2rij    = P4*em4x+P3*em3x+P2*em2x+P1*emx-2.*a*x;
    *d2rij /= (D*D)*(D*D);
  }
}

// called from ASRVsiteLikelihood with cl1=cl2
static void compute_proba_t3_cov(noeud nd, int OPTIMIZE_LENGTH, int cl1, int cl2){
/* Parameters:    branch lengths  :	 nd->l1 , nd->l2 		   */
/* 		  root location	  : 	 nd->l1/(nd->l1+nd->l2)		   */

  double rr1, rr2; 				/* relative rate */
  double p0, q; // p1, p2;
  int i,j,k,l,m;
  double elambdat[20]; // delambdat[20],ddelambdat[20];

  if(nd->v1==NULL) return;

	/* calculs des rr (=rij) */

  effective_length(nd->l1, cl1, cl2, covar3, &(nd->rr1[cl1][cl2]), &(nd->drr1[cl1][cl2]), &(nd->d2rr1[cl1][cl2]));
  effective_length(nd->l2, cl1, cl2, covar3, &(nd->rr2[cl1][cl2]), &(nd->drr2[cl1][cl2]), &(nd->d2rr2[cl1][cl2]));
  rr1=nd->rr1[cl1][cl2]; rr2=nd->rr2[cl1][cl2];

 /* proba. to branch 1 */
  for (k=0;k<20;k++) {
    elambdat[k] = exp(nd->l1*rr1*eigmat[k]);
//    delambdat[k] = (elambdat[k]*rr1*eigmat[k]);
//    ddelambdat[k] = (delambdat[k]*rr1*eigmat[k]);
  }

  for (m=0;m<20;m++) {
    double invfaa = 1.0/freqaa[m];
    for (l=0;l<20;l++) {
      p0=0.0; // p1=0.0; p2=0.0;
      for (k=0;k<20;k++) {
	q=probmat[k][m]*probmat[k][l];
	p0 += (q*elambdat[k]);
//	p1 += (q*delambdat[k]);
//	p2 += (q*ddelambdat[k]);

      }
      if(nd->alive1) {
	if(p0<0.) p0=0.;         // why negative happens ???
	Pind(nd->p1,gamma_nbcl1, cl1,m, cl2,l) = p0*invfaa;
//	for (int parameters=0;parameters<3;parameters++){
//	  nd->dp1[cl1][cl2][parameters][m][l]=p1/freqaa[m];
//	  nd->d2p1[cl1][cl2][parameters][m][l]=p2/freqaa[m];
//	}
      }
    }
  }

  if(Pind(nd->p1,gamma_nbcl1, cl1,0, cl2,1)<0. && verbose>=1){
    printf("l1=%f, rr1=%f\n", nd->l1, rr1);
    printf("cl1=%d, cl2=%d\n", cl1, cl2);
    printf("alive1: %d, alive2: %d\n", nd->alive1, nd->alive2);
  }

  /* proba. to branch 2 */
  for (k=0;k<20;k++) {
    elambdat[k] = exp(nd->l2*rr2*eigmat[k]);
//    delambdat[k] = (elambdat[k]*rr2*eigmat[k]);
//    ddelambdat[k] = (delambdat[k]*rr2*eigmat[k]);
  }

  for (m=0;m<20;m++) {
    double invfaa = 1.0/freqaa[m];
    for (l=0;l<20;l++) {
      p0=0.0; //p1=0.0; p2=0.0;
      for (k=0;k<20;k++) {
	q=probmat[k][m]*probmat[k][l];
	p0 += (q*elambdat[k]);
//	p1 += q*delambdat[k];
//	p2 += q*ddelambdat[k];
      }
      if(nd->alive2){
	if (p0<0.) p0=0.;
	Pind(nd->p2,gamma_nbcl1, cl1,m, cl2,l) = p0*invfaa;
// 	for (int parameters=0;parameters<3;parameters++){
// 	  nd->dp2[cl1][cl2][parameters][m][l]=p1/freqaa[m];
// 	  nd->d2p2[cl1][cl2][parameters][m][l]=p2/freqaa[m];
// 	}
      }
    }
  }


	/* IF UNRESOLVED NODE */

  if(nd->alive1==0){
    for(i=0;i<20;i++) for(j=0;j<20;j++)
      Pind(nd->p1,gamma_nbcl1, cl1,i, cl2,j) = (i==j)? 1 : 0;

//    for(i=0;i<20;i++) for(j=0;j<20;j++) for(int kk=0;kk<20;kk++)
//      nd->dp1[cl1][cl2][i][j][kk]=0;

//    for(i=0;i<20;i++) for(j=0;j<20;j++) for(int kk=0;kk<20;kk++)
//      nd->d2p1[cl1][cl2][i][j][kk]=0;
  }

  if(nd->alive2==0){
    for(i=0;i<20;i++) for(j=0;j<20;j++)
      Pind(nd->p2,gamma_nbcl1, cl1,i, cl2,j) = (i==j);

//     for(i=0;i<20;i++) for(j=0;j<20;j++) for(int kk=0;kk<20;kk++)
//       nd->dp2[cl1][cl2][i][j][kk]=0;

//     for(i=0;i<20;i++) for(j=0;j<20;j++) for(int kk=0;kk<20;kk++)
//       nd->d2p2[cl1][cl2][i][j][kk]=0;
  }

  compute_proba_t3_cov(nd->v1, OPTIMIZE_LENGTH, cl1, cl2);
  compute_proba_t3_cov(nd->v2, OPTIMIZE_LENGTH, cl1, cl2);
}



static void setDr(int rateClasses, double Dr[rateClasses][rateClasses]) {
  double pi_on = covar1/(covar1+covar2);
  int i;
  for (i=0; i<gamma_nbcl; i++)
    Dr[i][i] = gamma_clmean[i]/pi_on;
}

static void setG(int rateClasses, double G[rateClasses][rateClasses]) {
  int i,j;

  for (i=0; i<gamma_nbcl; i++)
    for (j=0; j<gamma_nbcl;j++) {
      if (i==j) G[i][j]=-covar2-covar3*(gamma_nbcl-1)/gamma_nbcl;
      else G[i][j] = covar3/gamma_nbcl;
    }

  for (i=0; i<gamma_nbcl; i++)
    for (j=gamma_nbcl; j<rateClasses; j++) {
      if (j==i+gamma_nbcl) G[i][j] = covar2;
      else G[i][j] = 0;
    }

  for (i=gamma_nbcl; i<rateClasses; i++)
    for (j=0; j<rateClasses; j++) {
      G[i][j] = 0;
      if (j==i-gamma_nbcl) G[i][j] = covar1;
      if (j==i) G[i][j] = -covar1;
    }
}

static void setG_Dr(int rateClasses, double G_Dr[rateClasses][rateClasses], double G[rateClasses][rateClasses], double Dr[rateClasses][rateClasses], int aminoIndex) {
  int i,j;
  for (i=0;i<rateClasses;i++)
    for (j=0;j<rateClasses;j++)
      G_Dr[i][j] = G[i][j]+eigmat[aminoIndex]*Dr[i][j];
}

/*one call to KroneckerProduct fills all rows of Kronecker and rateClasses columns*/
static void KroneckerProduct(int rateClasses, double Kronecker[20*rateClasses][20*rateClasses], double A[rateClasses][rateClasses], int aminoIndex)
{
  int ii, j, k;
  for (ii=0; ii<rateClasses; ii++)
    for (k=0; k<20; k++)
      for (j=0; j<rateClasses; j++)
	Kronecker[ii*20+k][aminoIndex*rateClasses+j] = A[ii][j]*eigenvect[k][aminoIndex];
}

/* generate a probability matrix G_Dr, based on covar1,2,3, gamma_shp (gamma_clmean),
 * and the (eigen decomposition of the) constant JTT matrix.
 * express our probability matrix as Vr * diag(rr) * Vr^-1
 * so we can exponentiate it by exponentiating the diagonal, rr.
 * Expand it with a Kronecker product with the eigenvectors of the JTT matrix
 * at the same time, so vr and invvr are 20*gamma_nbcl1 square. */
static int diagonalizeMatrix(double *vr, double *invvr, double *rr)
{
  int i, j;
  int m = 20;         // row size of amino acid subst. matrix
  int dim = m*gamma_nbcl1;
  double G[gamma_nbcl1][gamma_nbcl1]; // small matrices fit on the stack.
  double Dr[gamma_nbcl1][gamma_nbcl1];
  memset(Dr, 0, sizeof(Dr));
  double G_Dr[gamma_nbcl1][gamma_nbcl1];
  double rr_GDr[gamma_nbcl1];
  double vr_GDr[gamma_nbcl1][gamma_nbcl1];

  setG(gamma_nbcl1, G);		// Markov transition matrix of rates
  setDr(gamma_nbcl1, Dr);	// diagonal matrix of rates
  /* diagonalize JTT matrix - hard-coded in the aamatrix.c: eigmat and JTTeigenvect */
  for (i=0; i<m; i++) {
    setG_Dr(gamma_nbcl1, G_Dr, G, Dr, i); // symmetric iff covar1 == covar2. not worth optimizing for.
    callMyEigen(gamma_nbcl1, (double *)G_Dr, (double *)vr_GDr, rr_GDr);
    for (j=0; j<gamma_nbcl1; j++)
      rr[i*gamma_nbcl1+j] = rr_GDr[j];
    KroneckerProduct(gamma_nbcl1, (double (*)[])vr, vr_GDr, i);
  }
  return invmat(dim, vr, invvr);
}


/* rtop */
/* Compute transition probability matrix P from rate matrix R */
/* and branch length (time) t. */
/* lines (first dimension) of R should sum to zero:
for(i=0;i<dim;i++){
  R[i][i]=0.;
  for(j=0;j<dim;j++){
    if(j==i) continue;
    R[i][i]-=R[i][j];
  }
}
*/

static void rtop(int n, double t, double *P, double *vr, double *invvr,  double *rr)
{
  int i, j;
  double A[n*n];
  for(i=0;i<n;i++){ // do the diagonal * general matmul ourselves.
    double ediag = exp(rr[i]*t);
    for(j=0 ; j<n ; j++) A[pos(i,j,n)] = ediag*invvr[pos(i,j,n)];
  }
  matmat(n, n, vr, n, n, A, P);
}

static int P_rangecheck(int dim, double P[dim][dim])
{
  int m1, m2;
  for (m1=0 ; m1<dim ; m1++) for (m2=0 ; m2<dim ; m2++){
    if(!isgreaterequal(P[m1][m2],0.)){ /* isgreaterequal() returns false on NaN */
      if(!isgreaterequal(P[m1][m2],-0.0000001)) return m1*dim+m2; // prob. cannot be computed
      else P[m1][m2] = 0.;
    }
    else if(!islessequal(P[m1][m2], 1.)){
      if(!islessequal(P[m1][m2], 1.0000001)) return m1*dim+m2;
      else P[m1][m2] = 1.;
    }
  }
  return 0;
}


#define RANGECHECK_ERROR(bnum) do{\
  char *err=NULL; double badp = nd->p##bnum[retp]; \
  asprintf(&err, "procov: P_rangecheck found an error in branch %s:%.6g. P[%d][%d] =%s %.6g\n", \
	   nd->v##bnum->nom, nd->l##bnum, retp/dim, retp%dim, badp>1?" 1 +":"", badp>1? badp-1 : badp); \
  fputs(err, stderr); \
  if(loglevel>0) fputs(err, out); \
  free(err); \
  if(verbose>=4) pc_treeprint(findroot(nd, 10000)); \
}while(0);


static int compute_exact_probacov(noeud nd, double *vr, double *invvr, double *rr){
  int ret1, ret2, retp, dim = 20*gamma_nbcl1;
  if(nd->v1==NULL) return 1;
  ret1 = compute_exact_probacov(nd->v1, vr, invvr, rr);
  ret2 = compute_exact_probacov(nd->v2, vr, invvr, rr);

  rtop(dim, nd->l1, nd->p1, vr, invvr, rr); // bl1: compute P
  if((retp = P_rangecheck(dim, (double (*)[])nd->p1))){
    RANGECHECK_ERROR(1);
    return 0;
  }

  rtop(dim, nd->l2, nd->p2, vr, invvr, rr); // bl2
  if((retp = P_rangecheck(dim, (double (*)[])nd->p2))){
    RANGECHECK_ERROR(2);
    return 0;
  }

  return (ret1 && ret2);    /* return 1 is good */
}




// set l1, l2, l3 down the tree from s_tree->listbr.
void reset_bl(tree* s_tree, noeud nd)
{
  if(nd==NULL) return;
  if(nd!=root){
    double l = s_tree->listbr[nd->numerobr]->length;
    if(nd==root->v1)
      nd->l3=root->l1=l*s_tree->froot;
    else if(nd==root->v2)
      nd->l3=root->l2=l*(1.-s_tree->froot);
    else{
      nd->l3=l;
      if(nd==nd->v3->v1) nd->v3->l1=l;
      if(nd==nd->v3->v2) nd->v3->l2=l;
    }
  }

  reset_bl(s_tree, nd->v1);
  reset_bl(s_tree, nd->v2);
}

static void reset_tree(tree* s_tree)
{
  if(s_tree->which_var_param[1]>=0 || s_tree->which_var_param[0]>=0) reset_bl(s_tree, root);
  //  if(s_tree->which_var_param[2]>=0) set_node_comp(root, s_tree->GCanc, s_tree);
}

// impose upper and lower bounds on parameters,
// and flag which parameters hit the boundary
static void boundaries(tree* s_tree, int* bounded)
{
  int i, limit = s_tree->nbalivebr;

  for(i=0;i<s_tree->nb_var_param;i++)
    bounded[i]=0;

  if(s_tree->which_var_param[1]>=0){
    for(i=0;i<limit;i++)
      if(*(s_tree->var_param[s_tree->which_var_param[1]+i])<lmin){
        *(s_tree->var_param[s_tree->which_var_param[1]+i])=lmin;
        bounded[s_tree->which_var_param[1]+i]=1;
      }
  }

  static const double pmin[] = {frmin, lmin, covmin, covmin, covmin, pimin};
  static const double pmax[] = {frmax, lmax, covmax, covmax, covmax, pimax};
  for (i=0;i<=5;i++){
    if (i==1) continue; // special case
    if(s_tree->which_var_param[i]>=0){
      double tmp = *(s_tree->var_param[s_tree->which_var_param[i]]);
      tmp = min(pmax[i], max(pmin[i], tmp));
      bounded[i] = (tmp != *(s_tree->var_param[s_tree->which_var_param[i]]));
      *(s_tree->var_param[s_tree->which_var_param[i]]) = tmp;
    }
  }
}

#ifdef COMPILE_DEAD_CODE
static void resetdx(noeud nd, int param){
  int k, i, cl;

  if(nd->v1==NULL) return;

  for(cl=0;cl<gamma_nbcl;cl++){
    for(i=0;i<nd->nbpattern;i++){
      for(k=0;k<20;k++){
        nd->dx[cl][i][k]=0.;
      }
    }
  }
  resetdx(nd->v1, param);
  resetdx(nd->v2, param);
}

static void resetd2x(noeud nd){
  int k, i, cl;

  if(nd->v1==NULL) return;

  for(cl=0;cl<gamma_nbcl;cl++){
    for(i=0;i<nd->nbpattern;i++){
      for(k=0;k<20;k++){
        nd->d2x[cl][i][k]=0.;
      }
    }
  }

  resetd2x(nd->v1);
  resetd2x(nd->v2);

}
#endif // COMPILE_DEAD_CODE


void reset_like(tree* s_tree){
  int i, j, k, l, nbpattern;

  for(i=s_tree->nbseq;i<2*s_tree->nbseq-2;i++){
    nbpattern=s_tree->node[i]->nbpattern;
    for(j=0;j<nbpattern;j++){
      for(l=0;l<gamma_nbcl;l++){
	for(k=0;k<20;k++) {
	  s_tree->node[i]->x[j][l][k]=-1.;
	}
      }
    }
  }
}


static void dlike_dpi(tree* s_tree, double* dlike_s){
  int i;
  for(i=0;i<s_tree->lgseq;i++)
    dlike_s[i]=s_tree->like_s[i]-s_tree->likeasrv_s[i];
}


#ifdef COMPILE_DEAD_CODE
static void printMatrix(double **matrx, int nbparam1, int offset, FILE *out) {
  int i, j;

  for(i=offset;i<s_tree->nb_var_param;i++){
    switch(s_tree->param_nature[i]){
    case COVA: fprintf (out, "cov1\t"); break;
    case COVB: fprintf (out, "cov2\t"); break;
    case COVC: fprintf (out, "cov3\t"); break;
    case GaltierPI: fprintf (out, "GaltierPi\t"); break;
    }
  }

  if (options.compute.OPTIMIZE_GAMMA) fprintf(out, "alpha");
  fprintf (out, "\n");
  for (i=0; i<nbparam1; i++){
    for (j=0; j<nbparam1; j++)
      fprintf (out, "%.6f ", matrx[i][j]);
    fprintf (out, "\n");
  }
  fprintf (out, "\n");
}

static void covarianceMatrix(tree* s_tree, double *paramvect, double gamma_shp, FILE *out){
  int offset, ret, i, j, nbparam, nbparam1;
  double h, h2, h2square, *lkh1, l0, l1, l2, l3,l4,*remparam, **covarianceM, **fisherI;

  h = 0.001; h2 = 2*h; h2square=h2*h2;
  nbparam=0; offset = 0;

  /* we do not compute covariance matrix for edge length */
  for(i=0;i<s_tree->nb_var_param;i++){
    switch(s_tree->param_nature[i]){
    case ROOT: case LENGTH: offset ++; break;
    case COVA: case COVB: case COVC: case GaltierPI: nbparam++; break;
    }
  }

  if (options.compute.OPTIMIZE_GAMMA) nbparam1=nbparam+1;
  else nbparam1=nbparam;

  remparam = (double *) check_alloc(s_tree->nb_var_param, sizeof(double));

  lkh1 = (double *) check_alloc(nbparam1, sizeof(double));
  covarianceM = (double **)check_alloc(nbparam1, sizeof(double *));
  fisherI = (double **)check_alloc(nbparam1, sizeof(double *));

  for (i=0; i<nbparam1; i++) {
    covarianceM[i] = (double *) check_alloc(nbparam1, sizeof(double));
    fisherI[i] = (double *) check_alloc(nbparam1, sizeof(double));
  }

  for(i=0;i<s_tree->nb_var_param;i++)
    remparam[i]=*(s_tree->var_param[i]);

  /* store current log likelihoods */
  l0 = s_tree->lkh;

  for (i=0; i<nbparam; i++) {
    for (j=0; j<nbparam; j++) {
      remparam[i+offset] += h;
      remparam[j+offset] += h;
      l1 = -minus_log_likelihood(remparam);
      remparam[j+offset] -= h2;
      l2 = -minus_log_likelihood(remparam);
      remparam[i+offset] -= h2;
      l4 = -minus_log_likelihood(remparam);
      remparam[j+offset] += 2*h;
      l3 = -minus_log_likelihood(remparam);
      remparam[i+offset] += h; remparam[j+offset] -= h;

      fisherI[i][j] = -1*(l1-l2-l3+l4)/h2square;
    }
  }

  for (i=0; i<nbparam; i++)
    *(s_tree->var_param[i+offset])= remparam[i+offset];

  for(i=0;i<s_tree->nb_var_param;i++)
    remparam[i]=*(s_tree->var_param[i]);

  if (options.compute.OPTIMIZE_GAMMA) {
    for (i=0; i<nbparam; i++) {
      gamma_shp += h;
      DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
      remparam[i+offset] += h;
      l1 = -minus_log_likelihood(remparam);

      remparam[i+offset] -= h2;
      l2 = -minus_log_likelihood(remparam);

      gamma_shp -= h2;
      DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
      l4 = -minus_log_likelihood(remparam);

      remparam[i+offset] += h2;
      l3 = -minus_log_likelihood(remparam);

      fisherI[i][nbparam] = -1*(l1-l2-l3+l4)/h2square;
      fisherI[nbparam][i] = fisherI[i][nbparam];

      gamma_shp += h;
      remparam[i+offset] -= h;
      *(s_tree->var_param[i+offset])= remparam[i+offset];
    }

    /* get diagonal entry for alpha in the fisher information matrix */
    gamma_shp += h2;
    DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
    l1 = -minus_log_likelihood(remparam);
    gamma_shp -= 2*h2;
    DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
    l4 = -minus_log_likelihood(remparam);
    fisherI[nbparam][nbparam] = -1*(l1+l4-2*l0)/h2square;
    gamma_shp += h2;
    DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
  }

  s_tree->lkh = l0;

  fprintf(out, "fisher information matrix\n");
  printMatrix(fisherI, nbparam1, offset, out);

  /* inverse fisher information matrix to get covariance matrix
     if ret = 0 if the fisher information matrix is singular */
  ret = invmat(fisherI, nbparam1, covarianceM);
  if (ret == 0) fprintf (out, "Singular fisher information matrix! Covariance matrix cannot be computed.\n");
  else {
    fprintf(out, "covariance matrix\n");
    printMatrix (covarianceM, nbparam1, offset, out);
  }

  for (i=0; i<nbparam1; i++) {free (covarianceM[i]); free (fisherI[i]);}
  free (covarianceM); free(fisherI); free(lkh1); free(remparam);
}
#endif // COMPILE_DEAD_CODE


/* Computes first and second derivatives of log likelihood conditional */
/* to rate class (class). */

static void derivatives(tree* s_tree, double *paramvect){
  int i, j;
  double like, *like_s,*likeasrv_s, dlike;
  double *remparam;
  double l0, l1, l2;
  int nbparam = s_tree->nb_var_param;
  int lgseq = s_tree->lgseq;
  double h = 0.00001, hsquare = h*h;

  //resetdx(root, 0);
  // store likelihood and site likelihood for the tree of current parameters
  l0 = s_tree->lkh;
  like_s = (double *) check_alloc(lgseq, sizeof(double));
  likeasrv_s = (double *) check_alloc(lgseq, sizeof(double));
  for (i=0; i<lgseq; i++){
    like_s[i] = s_tree->like_s[i];
    likeasrv_s[i] = s_tree->likeasrv_s[i];
  }

  remparam = (double *) check_alloc(nbparam, sizeof(double));
  for(i=0;i<nbparam;i++)
    remparam[i]=*(s_tree->var_param[i]);

  for (i=0; i<nbparam; i++) {
    if (s_tree->param_nature[i] != GaltierPI) { // we do analytical deriv for PI_galtier
      remparam[i] += h;          // x+h
      l1 = -minus_log_likelihood(remparam);  // f(x+h)

      remparam[i] += h;     // x+2h
      l2 = -minus_log_likelihood(remparam);     // f(x+2h)

      remparam[i] -= 2*h;    // restore the original parameter
      *(s_tree->var_param[i])= remparam[i];

      s_tree->deriv[i] = (l1-l0)/h;
      s_tree->deriv2[i] = (l2+l0-2*l1)/hsquare;
    }
    else {
      dlike_dpi(s_tree, s_tree->dlike_s[i]);
      s_tree->deriv[i]=s_tree->deriv2[i]=0.;
      for(j=0;j<lgseq;j++){
	like=pi*like_s[j]+(1-pi)*likeasrv_s[j];
	dlike=s_tree->dlike_s[i][j];

	s_tree->deriv[i]+=s_tree->weight[j]*dlike/like;
	s_tree->deriv2[i]+=s_tree->weight[j]*(-(dlike*dlike/(like*like)));
      }
    }
  }

  //  restore tree likelihood and site likelihoods to the original value before calling the function
  s_tree->lkh = l0;
  for (i=0; i<lgseq; i++){
    s_tree->like_s[i] = like_s[i];
    s_tree->likeasrv_s[i] = likeasrv_s[i];
  }

  free(remparam);
  free(like_s);free(likeasrv_s);
}

#ifdef COMPILE_DEAD_CODE
/* like_node_2sites */
/* Compute joint likelihood of the current two sites for node nd */

static void like_node_2sites(noeud nd){

  int i, j, i1, j1, cl, cl1, deb, fin;
  double s1, s2, t1, t2;

  if(nd->v1==NULL){
    return;
  }

  like_node_2sites(nd->v1);
  like_node_2sites(nd->v2);

  for(cl=0;cl<gamma_nbcl;cl++){
    if(covar>0.){deb=0; fin=gamma_nbcl;} else {deb=cl; fin=cl+1;}
    for(i=0;i<20;i++){
      for(j=0;j<20;j++){
        s1=s2=0.;
        for(cl1=deb;cl1<fin;cl1++){
          for(i1=0;i1<20;i1++){
            for(j1=0;j1<20;j1++){
              t1=nd->v1->x2sites[cl1][i1][j1]*nd->p1[cl][cl1][i][i1]*nd->p1[cl][cl1][j][j1];
              t2=nd->v2->x2sites[cl1][i1][j1]*nd->p2[cl][cl1][i][i1]*nd->p2[cl][cl1][j][j1];
              s1+=t1;
              s2+=t2;
            }
          }
        }
        nd->x2sites[cl][i][j]=s1*s2;
        if(nd->x2sites[cl][i][j]<0.){
          printf("i:%d , j:%d , nd->x2sites[cl][i][j]:%.20f\n", i, j, nd->x2sites[cl][i][j]);
          exit(0);
        }
      }
    }
  }
}



/* init_x2sites */
/* Initialize x2sites at tips for sites s1 and s2 */

static void init_x2sites(noeud nd, int s1, int s2){

  int j, k, l, p1, p2;
  int aa, aa1;

  if(nd->v1){ init_x2sites(nd->v1, s1, s2); init_x2sites(nd->v2, s1 ,s2); return;}

  p1=setpattern_site(nd, s1);
  p2=setpattern_site(nd, s2);

  printf("p1=%d\tp2=%d\n",p1,p2);
  for(j=0;j<gamma_nbcl;j++){
    for(k=0;k<20;k++){ /* 20 possible states at site 1 */
      for(l=0;l<20;l++){ /* 20 possible states at site 2 */
	if(k==p1 && l==p2) nd->x2sites[j][k][l]=1.;
	else nd->x2sites[j][k][l]=0.;
      }
    }
  }

  // special cases
  if (p1==20) {    // B
    if (l==p2) {nd->x2sites[j][2][l]=1.; nd->x2sites[j][3][l]=1.;}
    if (p2==20) {
      nd->x2sites[j][2][2]=1.; nd->x2sites[j][2][3]=1.;
      nd->x2sites[j][3][2]=1.; nd->x2sites[j][3][3]=1.;
    }
    if (p2==21) {
      nd->x2sites[j][2][5]=1.; nd->x2sites[j][2][6]=1.;
      nd->x2sites[j][3][5]=1.; nd->x2sites[j][3][6]=1.;
    }
    if (p2==22) {
      for (aa=0;aa<20;aa++) {nd->x2sites[j][2][aa]=1.; nd->x2sites[j][3][aa]=1.;}
    }
  }

  if (p1==21) {     // Z
    if (l==p2) {nd->x2sites[j][5][l]=1.; nd->x2sites[j][6][l]=1.;}
    if (p2==20) {
      nd->x2sites[j][5][2]=1.; nd->x2sites[j][5][3]=1.;
      nd->x2sites[j][6][2]=1.; nd->x2sites[j][6][3]=1.;
    }
    if (p2==21) {
      nd->x2sites[j][5][5]=1.; nd->x2sites[j][5][6]=1.;
      nd->x2sites[j][6][5]=1.; nd->x2sites[j][6][6]=1.;
    }
    if (p2==22) {
      for (aa=0;aa<20;aa++) {nd->x2sites[j][5][aa]=1.; nd->x2sites[j][6][aa]=1.;}
    }
  }

  if (p1==22) {   // gaps, unknnwn residues
    if (l==p2) {
      for (aa=0;aa<20;aa++)
	nd->x2sites[j][aa][l]=1.;
    }
    if (p2==20) {
      for (aa=0;aa<20;aa++) {
	nd->x2sites[j][aa][2]=1.; nd->x2sites[j][aa][3]=1.;
      }
    }
    if (p2==21) {
      for (aa=0;aa<20;aa++) {
	nd->x2sites[j][aa][5]=1.; nd->x2sites[j][aa][6]=1.;
      }
    }
    if (p2==22) {
      for (aa=0;aa<20;aa++)
	for (aa1=0;aa1<20;aa1++)
	  nd->x2sites[j][aa][aa1]=1.;
    }
  }


  if (p2==20) {    // B
    if (k==p1) {nd->x2sites[j][k][2]=1.; nd->x2sites[j][k][3]=1.;}
    if (p1==21) {
      nd->x2sites[j][5][2]=1.; nd->x2sites[j][5][3]=1.;
      nd->x2sites[j][6][2]=1.; nd->x2sites[j][6][3]=1.;
    }
    if (p1==22) {
      for (aa=0;aa<20;aa++) {nd->x2sites[j][aa][2]=1.; nd->x2sites[j][aa][3]=1.;}
    }
  }

  if (p2==21) {     // Z
    if (k==p1) {nd->x2sites[j][k][5]=1.; nd->x2sites[j][k][6]=1.;}
    if (p1==20) {
      nd->x2sites[j][2][5]=1.; nd->x2sites[j][2][6]=1.;
      nd->x2sites[j][3][5]=1.; nd->x2sites[j][3][6]=1.;
    }
    if (p1==22) {
      for (aa=0;aa<20;aa++) {nd->x2sites[j][aa][5]=1.; nd->x2sites[j][aa][6]=1.;}
    }
  }

  if (p2==22) {   // gaps, unknnwn residues
    if (k==p1) {
      for (aa=0;aa<20;aa++)
	nd->x2sites[j][k][aa]=1.;
    }
    if (p1==20) {
      for (aa=0;aa<20;aa++) {
	nd->x2sites[j][2][aa]=1.; nd->x2sites[j][3][aa]=1.;
      }
    }
    if (p1==21) {
      for (aa=0;aa<20;aa++) {
	nd->x2sites[j][5][aa]=1.; nd->x2sites[j][6][aa]=1.;
      }
    }

  }

}


/* like_2sites */
/* Compute the joint likelihood of sites s1 and s2 assuming */
/* common changes of rate class */
/* note: patterns and patternsons are not used for this calculation */

static double like_2sites(tree* s_tree, int s1, int s2){
  int i, j, k;
  double like;

  init_x2sites(root, s1, s2);

  like_node_2sites(root);

  like=0.;
  for(k=0;k<gamma_nbcl;k++){
    for (i=0;i<20;i++)
      for (j=0;j<20;j++)
	like+=root->x2sites[k][i][j]*freqaa[i]*freqaa[j];
  }

  if(like<0.){
    printf("x2sites %d %d: like <0\n", s1, s2);
    return 1.;
  }

  like/=(gamma_nbcl*gamma_nbcl);
  return like;
}
#endif // COMPILE_DEAD_CODE


static int ASRVsiteLikelihood (tree* s_tree)
{
  int i, j, k;

  //  if(covar3>=0. && pi>=0.){   /* remove the condition 23/03/2006*/
  if (1) {
    double memcovar = covar3;
    covar3 = -1;

    for(k=0;k<gamma_nbcl;k++)
      compute_proba_t3_cov(root, options.compute.OPTIMIZE_LENGTH, k, k);

    ASRVlike_node(root);

    for(k=0;k<gamma_nbcl;k++){
      for(i=0;i<s_tree->lgseq;i++){
	s_tree->classlike_s[k][i]=0;
	for(j=0;j<20;j++) s_tree->classlike_s[k][i] +=root->x[i][k][j]*freqaa[j];
        if(s_tree->classlike_s[k][i]<0.){
          fprintf(stderr, "procov: error in %s: site : %d , argument=%e\n", __func__, i, s_tree->classlike_s[k][i]);
          fprintf(stderr, "%d : %e %e %e %e\n", i, root->x[i][k][0], root->x[i][k][1], root->x[i][k][2], root->x[i][k][3]);
          return 1;   // site likehood cannot be computed
        }
      }
    }

    for(i=0;i<s_tree->lgseq;i++){
      s_tree->likeasrv_s[i]=0.;
      for(k=0;k<gamma_nbcl;k++)
        s_tree->likeasrv_s[i]+=s_tree->classlike_s[k][i];
      s_tree->likeasrv_s[i]/=gamma_nbcl;
    }
    covar3=memcovar;
  }
  return 0;     // site likehood computed
}


static int CovarionSiteLikelihood (tree* s_tree) {
  int ret, i, j, k;
  int m = 20;         // row size of amino acid subst. matrix
  int dim = m*gamma_nbcl1;
  static double last_cov1=-1, last_cov2=-1, last_cov3=-1, last_gamma_shp=-1;
  static double *rr, *vr, *invvr;
  if(!rr){
    rr = check_alloc(dim, sizeof(double));
    vr = check_alloc(dim*dim, sizeof(*vr));
    invvr = check_alloc(dim*dim, sizeof(*invvr));
  }

  // we tend to get called a few times in a row with the same covar parameters
  if( last_cov1 == covar1 && last_cov2 == covar2 && last_cov3 == covar3 && last_gamma_shp == gamma_shp )
    ret = 1;
  else{
    last_cov1 = covar1; last_cov2 = covar2; last_cov3 = covar3; last_gamma_shp = gamma_shp;
    ret = diagonalizeMatrix(vr, invvr, rr);
  }

  if (ret == 0) {
    fputs("procov: error: singular eigen vector matrix\n", stderr);
    if (loglevel>=1) fputs("procov: error: singular eigen vector matrix\n", out);
    return 1;
  }

  ret = compute_exact_probacov(root, vr, invvr, rr);
  if (ret == 0) {
    char *err=NULL;
    asprintf(&err, "procov: error: transition probability cannot be computed.  continuing anyway...\n"
	     "pi: %8.6f\tgamma_nbcl1: %d\tgamma_shp: %.6f\n"
	     "cov1: %.12f\tcov2: %.12f\tcov3: %.12f\n",
	     pi, gamma_nbcl1, gamma_shp,
	     covar1,covar2,covar3);
    fputs(err, stderr);
    if (loglevel>=1) fputs(err, out);
    free(err);
    // return 1;
  }
//  free(rr); free(vr); free(invvr);

  Covarionlike_node(root);

  for(k=0;k<gamma_nbcl1;k++){
    for(i=0;i<s_tree->lgseq;i++){
      s_tree->classlike_s[k][i]=0;
      for(j=0;j<20;j++) s_tree->classlike_s[k][i] +=root->x[i][k][j]*freqaa[j];

      if(s_tree->classlike_s[k][i]<0.){
        fprintf(stderr, "procov: error in %s: site : %d , argument=%e\n", __func__, i, s_tree->classlike_s[k][i]);
        fprintf(stderr, "%d : %e %e %e %e\n", i, root->x[i][k][0], root->x[i][k][1], root->x[i][k][2], root->x[i][k][3]);
        return 1;          // site likehood cannot be computed
      }
    }
  }

  double pi_on = covar1/(covar1+covar2);
  for(i=0;i<s_tree->lgseq;i++){
    s_tree->like_s[i]=0.;
    for(k=0;k<gamma_nbcl;k++)
      s_tree->like_s[i]+=s_tree->classlike_s[k][i];

    s_tree->like_s[i] *= pi_on;

    for(k=gamma_nbcl;k<gamma_nbcl1;k++)
      s_tree->like_s[i]+=s_tree->classlike_s[k][i]*(1-pi_on);
    s_tree->like_s[i]/=gamma_nbcl;

  }

  return 0;     // site likehood computed
}


/* minus_log_likelihood computes minus the log-likelihood of tree s_tree (global) */
/* for parameters paramvprov and Gamma distribution given by */
/* gamma_nbcl and gamma_clmean (global) */

static double minus_log_likelihood(double *paramprov)
{
  double l;
  int i, nbparam = s_tree->nb_var_param;
  for(i=0;i<nbparam;i++) {
    *(s_tree->var_param[i])=paramprov[i];
  }
  int *bounded = check_alloc(nbparam,sizeof(int));
  boundaries(s_tree, bounded); /* let the parameters bounded in the preset lower
				  and upper bounds: this only works for those parameters
				  set to be optimized, as defined in the option file */
  free(bounded);
  reset_tree(s_tree);

  if (!(pi == 1. && !options.compute.OPTIMIZE_PI))
    ASRVsiteLikelihood (s_tree);  /* compute ASRV for non Huelsenbeck models */

  if (pi>0.) {       /* add a condition for doing covarion  23/03/2006*/
    int ret = CovarionSiteLikelihood (s_tree);
    if (ret) return 1.;     // transition prob. and log likelihood cannot be computed
  }

  l = 0;
  if (pi>0.) {
    if (pi==1. && !options.compute.OPTIMIZE_PI) {  /* Huelsenbeck's model*/
      for(i=0;i<s_tree->lgseq;i++)
	l+=log(s_tree->like_s[i])*s_tree->weight[i];
    }
    else {      /* Galtier or General models*/
      for(i=0;i<s_tree->lgseq;i++)
	l+=log(pi*s_tree->like_s[i]+(1-pi)*s_tree->likeasrv_s[i])*s_tree->weight[i];
    }
  }
  else {       /* ASRV model only */
    for(i=0;i<s_tree->lgseq;i++)
      l+=log(s_tree->likeasrv_s[i])*s_tree->weight[i];  /* change like_s to likeasrv_s. 23/03/2006 */
  }

  if(l>0.) {
    fputs("procov: error: log likelihood greater than 0! parameters :\n", stderr);
    for(i=0;i<nbparam;i++)
      fprintf(stderr, "%.10f\n", *(s_tree->var_param[i]));
    exit(3);
  }

  return -l;
}


#ifdef COMPILE_DEAD_CODE
static void set_proba_zero(noeud nd){

  int i, j, cl1, cl2, p;
  if(nd==NULL) return;

  for(cl1=0;cl1<gamma_nbcl;cl1++){
    for(cl2=0;cl2<gamma_nbcl;cl2++){
      for(i=0;i<20;i++){
        for(j=0;j<20;j++){
          nd->p1[cl1][cl2][i][j]=0.;
          nd->p2[cl1][cl2][i][j]=0.;
          for(p=0;p<5;p++){
            nd->dp1[cl1][cl2][p][i][j]=0.;
            nd->d2p1[cl1][cl2][p][i][j]=0.;
            nd->dp2[cl1][cl2][p][i][j]=0.;
            nd->d2p2[cl1][cl2][p][i][j]=0.;
          }
        }
      }
    }
  }

  set_proba_zero(nd->v1);
  set_proba_zero(nd->v2);
}


static double true_length(noeud nd){

  double mylg, teta, lat, lcg, glat, glcg, coeff;
  noeud v3;
  int i;
  double alpha=2;               // temporary setting
  if(nd==root) return -1.;

  coeff=0.;
  v3=nd->v3;
  mylg=nd->l3;
  teta=nd->ceq+nd->geq;
  lat=(1+alpha*teta)/2.;
  lcg=(1+alpha*(1-teta))/2.;

  for(i=0;i<gamma_nbcl;i++){
    glat=gamma_clmean[i]*lat;
    glcg=gamma_clmean[i]*lcg;

    coeff+= v3->a*glat + v3->c*glcg + v3->g*glcg + v3->t*glat;
  }

  return mylg*coeff/gamma_nbcl;
}

static double my_length(noeud nd, double alpha){

  double truelg, coeff, teta, lat, lcg, glat, glcg;
  noeud v3;
  int i;

  if(nd==root) return -1.;

  coeff=0.;
  v3=nd->v3;
  truelg=nd->l3;
  teta=nd->ceq+nd->geq;
  lat=(1+alpha*teta)/2.;
  lcg=(1+alpha*(1-teta))/2.;

  for(i=0;i<gamma_nbcl;i++){
    glat=gamma_clmean[i]*lat;
    glcg=gamma_clmean[i]*lcg;

    coeff+= nd->aeq*glat + nd->ceq*glcg + nd->geq*glcg + nd->teq*glat;
/* note: equil base comp are used here instead of node base comp */
/* because this (approximate) calculation occurs at the beginning */
/* of the prog, i.e. when node base comp are unknown */
  }

  return truelg*gamma_nbcl/coeff;
}

static void mylg_to_truelg(noeud nd){

  double tl;
  noeud v3;

  if(nd==NULL) return;

  tl=true_length(nd);

  if(tl>=0.){
    nd->l3=tl;
    v3=nd->v3;
    if(nd==v3->v1) v3->l1=tl; else v3->l2=tl;
  }

  mylg_to_truelg(nd->v1);
  mylg_to_truelg(nd->v2);
}

static void truelg_to_mylg(noeud nd, double alpha){

  double ml;
  noeud v3;

  if(nd==NULL) return;

  ml=my_length(nd, alpha);

  if(ml>=0.){
    nd->l3=ml;
    v3=nd->v3;
    if(nd==v3->v1) v3->l1=ml; else v3->l2=ml;
  }


  truelg_to_mylg(nd->v1, alpha);
  truelg_to_mylg(nd->v2, alpha);

}
#endif // COMPILE_DEAD_CODE

static double simplex(double* initpoint, double bound, double* finalvect)
{
  int i, j, nbiter, min, nbparam = s_tree->nb_var_param;
  double ran, ret;
  double *tempoint = check_alloc(nbparam, sizeof(double));
  double *pointlike = check_alloc(nbparam+2, sizeof(double));
  double **point = check_alloc(nbparam+2, sizeof(double*));
  for(i=0;i<=nbparam+1;i++)
    point[i]=(double*)check_alloc(nbparam+1, sizeof(double));
  double **contraintes = check_alloc(nbparam+1, sizeof(double*));
  for(i=1;i<=nbparam;i++){
    contraintes[i]=(double*)check_alloc(2, sizeof(double));
    contraintes[i][0]=min_value[s_tree->param_nature[i-1]];
    contraintes[i][1]=max_value[s_tree->param_nature[i-1]];
  }

/* choose N+1 initial points around initial values given as argument */
/* WARNING : arrays start from 1 */
  for(i=2;i<=nbparam+1;i++){
    for(j=1;j<=nbparam;j++){
      ran=drand48();	/*  0<ran<1  */
      ran/=5;		/*  0<ran<0.2  */
      ran+=0.9;		/* 0.9<ran<1.1 */
      point[i][j]=initpoint[j-1]*ran;
      if(point[i][j]<contraintes[j][0]) point[i][j]=contraintes[j][0];
      else if(point[i][j]>contraintes[j][1]) point[i][j]=contraintes[j][1];
    }
  }
  for(j=1;j<=nbparam;j++)
    point[1][j]=initpoint[j-1];

/* compute the likelihoods of each initial point */
  for(i=1;i<=nbparam+1;i++){
    for (j=0; j<nbparam; j++)
      tempoint[j] = point[i][j];
    //    pointlike[i]=minus_log_likelihood(point[i]);
    pointlike[i]=minus_log_likelihood(tempoint);
  }

/* simplex algorithm */
  amoeba_cont(point, pointlike, nbparam, bound, minus_log_likelihood, &nbiter, contraintes);

/* output best point */
  min=1;
  for(i=2;i<=nbparam+1;i++){
    if(pointlike[i]<pointlike[min])
      min=i;
  }
  for(i=0;i<nbparam;i++) finalvect[i]=point[min][i+1];

  ret=-pointlike[min];

  for(i=1;i<=nbparam;i++)
    free(contraintes[i]);
  free(contraintes);
  for(i=0;i<=nbparam+1;i++)
    free(point[i]);
  free(point);
  free(pointlike);
  free(tempoint);
  return ret;
}


#ifdef COMPILE_DEAD_CODE
typedef struct upgmanoeud{
	struct upgmanoeud *v1, *v2, *v3;	/* neighbors */
	double l1, l2, l3;
        double b1, b2, b3;
        int num;
        int alive;
        double depth;
	char* nom;
} *upgmanoeud;

static void set_alive_0(upgmanoeud nd){

  if(nd->v1==NULL){
    nd->alive=0;
    return;
  }

  set_alive_0(nd->v1);
  set_alive_0(nd->v2);
}

static void set_tiplg_min(upgmanoeud nd, double tiplgmin){

  if(nd->v1==NULL){
    nd->l3-=tiplgmin;
    nd->v3->l1=nd->v3->l2=nd->l3;
    return;
  }

  set_tiplg_min(nd->v1, tiplgmin);
  set_tiplg_min(nd->v2, tiplgmin);
}

static char* labelled_upgma(int nb, char** nom, double** dist, double *label){

  int i, j, j1, j2, k1, k2, iter, nbn, nbm1, ndnum;
  double dmin, d1, d2, tiplgmin;
  char* ctree;
  upgmanoeud* stree;


  /* alloc nodes */

  nbn=2*nb-1;
  stree=check_alloc(nbn, sizeof(upgmanoeud));

  /* set tip nodes */

  for(i=0;i<nb;i++){
    stree[i]=check_alloc(1, sizeof(struct upgmanoeud));
    stree[i]->v1=stree[i]->v2=NULL;
    stree[i]->l1=stree[i]->l2=-1;
    stree[i]->num=i;
    stree[i]->alive=1;
    stree[i]->depth=0.;
    stree[i]->nom=nom[i];
    stree[i]->b3=label[i];
  }

  /* main loop */

  nbm1=nb-1;
  ndnum=nb;
  for(iter=0;iter<nbm1;iter++){

	/* find min dist */
    dmin=DBL_MAX;
    for(i=0;i<nb;i++){
      if(stree[i]->alive==0) continue;
      for(j=i+1;j<nb;j++){
        if(stree[j]->alive==0) continue;
        if(dist[i][j]<dmin){
          dmin=dist[i][j];
          j1=i; j2=j;
        }
      }
    }

	/* find nodes to cluster */
    for(i=0;i<ndnum;i++)
      if(stree[i]->num==j1){k1=i; break;}
    for(i=0;i<ndnum;i++)
      if(stree[i]->num==j2){k2=i; break;}

	/* cluster */
    stree[ndnum]=check_alloc(1, sizeof(struct upgmanoeud));
    stree[ndnum]->v1=stree[k1];
    stree[ndnum]->v2=stree[k2];
    stree[k1]->v3=stree[ndnum];
    stree[k2]->v3=stree[ndnum];
    stree[ndnum]->l1=stree[k1]->l3=dmin/2.-stree[k1]->depth;
    stree[ndnum]->l2=stree[k2]->l3=dmin/2.-stree[k2]->depth;
    if(iter==0) tiplgmin=stree[ndnum]->l1;
    stree[ndnum]->depth=dmin/2.;
    stree[ndnum]->num=stree[k1]->num;
    stree[ndnum]->b3=(stree[k1]->b3+stree[k2]->b3)/2.;
    stree[k1]->num=-stree[k1]->num-1;
    stree[k2]->num=-stree[k2]->num-1;
    set_alive_0(stree[k2]);
    ndnum++;

	/* update distances */

    for(i=0;i<nb;i++){
      if(i<j1) d1=dist[i][j1];
      else d1=dist[j1][i];
      if(i<j2) d2=dist[i][j2];
      else d2=dist[j2][i];
      if(i<j1) dist[i][j1]=(d1+d2)/2.;
      if(i>j1) dist[j1][i]=(d1+d2)/2.;
    }

  }

  /* struct to string */

  ctree=check_alloc(nb*100, sizeof(char));
  upgmaroot=stree[ndnum-1];
  upgmaroot->v3=NULL;
  set_tiplg_min(upgmaroot, tiplgmin);
  upgma_stoc(stree, 1, nb, ctree, 0);

  return ctree;

}


/* output_covar_pattern */
/* Compute and print into file outcovfile : */
/* - the "variation index" of each site (contribution to the total likelihood */
/*   of scenarii involving at least one change of rate class ) */
/* - the "covariation index" of each pair of site (likelihood conditional */
/*   on simultaneous chanegs of rate class divided by unconstrained likelihood) */

static void output_covar_pattern(tree* s_tree, double lcovar, double* bestvect, int optcov, FILE *outcov){

  int i, j, k, ii, jj, lgseq, truelgseq, *pattnumber;
  int maxsitepername=5;
  double lnoch, lcov, *varind, **covarind;
  char **name, **longname, **finlongname, *prov, *covartree;
  //  FILE* outcov;

  printf("computing variation and covariation indices...\n");

  //  outcov=fopen("covarfile", "w");
  if(outcov==NULL) return;
  lgseq=s_tree->lgseq;
  truelgseq=s_tree->weightsum;

  varind=check_alloc(lgseq, sizeof(double));
  covarind=check_alloc(lgseq, sizeof(double*));
  for(k=0;k<lgseq;k++) covarind[k]=check_alloc(lgseq, sizeof(double));
  pattnumber=check_alloc(lgseq, sizeof(int));
  longname=check_alloc(lgseq, sizeof(char*));
  finlongname=check_alloc(lgseq, sizeof(char*));
  name=check_alloc(lgseq, sizeof(char*));
  for(i=0;i<lgseq;i++) name[i]=check_alloc(maxsitepername*(log10(lgseq)+2), sizeof(char));


  minus_log_likelihood(bestvect-1);

  /* store current likelihoods */

  for(i=0;i<lgseq;i++){
    varind[i]=s_tree->like_s[i];
    for(j=i;j<lgseq;j++) covarind[i][j]=s_tree->like_s[i]*s_tree->like_s[j];
  }

  /* variation indices */

  covar=-1.;
  if(optcov) bestvect[s_tree->which_var_param[2]]=-1.;  // let covarion rate = -1.
  //  log_like_total(s_tree);     // temporary disable it

  for(i=0;i<lgseq;i++){
    lnoch=s_tree->like_s[i];
    varind[i]=1.-(lnoch/varind[i]);
  }

  covar=lcovar;
  if(optcov) bestvect[s_tree->which_var_param[2]]=lcovar;   // change covar to the original value

  /* covariation indices */

  for(i=0;i<lgseq;i++){
    for(j=i;j<lgseq;j++){
      lcov=like_2sites(s_tree, i, j);
      covarind[i][j]=lcov/covarind[i][j];
    }
  }

  /* my change
  fprintf(outcov, "\nVariation index for each site:\n\n");
  for (ii=0; ii<lgseq; ii++)
    fprintf (outcov, "%d\t%.4e\n",ii+1, varind[ii]);

  fprintf (outcov, "\ncovariation indices\n");
  for (i=0; i<lgseq; i++)
    for (ii=0; ii<lgseq; ii++)
      fprintf (outcov, "%d\t%d\t%.4e\n",i+1, ii+1, covarind[i][ii]);
  */

  /* UPGMA tree for covariation indices */

	/* set names of site groups */
  ii=0;
  for(i=0;i<truelgseq;i++){
    j=s_tree->sitetopatt[i];
    k=pattnumber[j];
    pattnumber[j]++;
    if(k<maxsitepername){
      prov=name[j];
      while(*prov) prov++;
      if(prov==name[j]) sprintf(prov, "%d", i+1);
      else sprintf(prov, "_%d", i+1);
    }
    if(k==maxsitepername){
      longname[ii]=check_alloc((log10(lgseq)+1)*lgseq, sizeof(char));
      sprintf(longname[ii], "set%d: %s_%d", ii+1, name[j], i+1);
      sprintf(name[j], "set%d", ii+1);
      finlongname[ii]=longname[ii];
      while(*(finlongname[ii])) finlongname[ii]++;
      ii++;
    }
    if(k>maxsitepername){
      prov=name[j];
      sscanf(prov, "set%d", &jj);
      jj--;
      sprintf(finlongname[jj], "_%d", i+1);
      while(*(finlongname[jj])) finlongname[jj]++;
    }
  }

	/* compute distances */

  for(i=0;i<lgseq;i++)
    for(j=i+1;j<lgseq;j++)
       covarind[i][j]=1.-covarind[i][j];


	/* build tree */

  covartree=labelled_upgma(lgseq, name, covarind, varind);



	/* output tree */

  fprintf(outcov, "\nCovariation gaph:\n");
  fprintf(outcov, "%s\n", covartree);
  for(i=0;i<ii;i++){
    fprintf(outcov, "%s\n", longname[i]);
  }

  /* output variation indices */

  fprintf(outcov, "\nVariation index for each site:\n\n");

  ii=0;
  while(1){
    for(j=0;j<10;j++){
      for(k=0;k<5;k++){
        i=ii*50+j*5+k;
        if(i>=truelgseq) break;
        fprintf(outcov, "%d\t%.4e\n", i+1, varind[s_tree->sitetopatt[i]]);
      }
      if(i>=truelgseq) break;
    }
    if(i>=truelgseq) break;
    ii++;
  }

  //  fclose(outcov);
}
#endif // COMPILE_DEAD_CODE

static double compute_treeLength(noeud nd) {
  double tl1,tl2;

  if (nd->v1)
    tl1=compute_treeLength(nd->v1)+nd->l1;
  else tl1=0.;

  if (nd->v2)
    tl2=compute_treeLength(nd->v2)+nd->l2;
  else tl2=0.;

  return (tl1+tl2);
}

static void getTreelength(double *oriTreeLen, double *treeLen)
{
  int i; double sum0 = 0;
  for(i=0;i<2*s_tree->nbseq-3;i++) sum0 += s_tree->lgbr_init[i];
  *oriTreeLen = sum0;

  if(options.compute.OPTIMIZE_LENGTH)
    *treeLen = compute_treeLength(root);
  else *treeLen = *oriTreeLen;
}



double compute(tree* s_tree, char *status)
{
  int i, j, ii, nbseq, lgseq, nbparam, firstcomp, nbreduction, best_gamma;
  int OPTIMIZE, nbiter, *bounded, do_simplex=0, do_simplex_now=0;
  double rem, bound, *toadd, totallength, newtotallength, gamma_l, best_gamma_l;
  double llike, maxllike, *paramvect, *remvect, exact_llike;
  double prec, lll, oriTreeLen, treeLen;
  double* bestvect, *finalvect;
  char *ctmp;
  int den;
  double l1, l_1, delta, d_gamma, d2_gamma, toadd_gamma, old_gamma_shp;
  static FILE *outcov=NULL, *siteLKHfile;
  char *ctree;
  maxllike = 0; // keep gcc happy

  nbseq=s_tree->nbseq;
  lgseq=s_tree->lgseq;
  nbparam=s_tree->nb_var_param;
 
  prec=options.converge.PRECISION;
  
  OPTIMIZE = options.compute.OPTIMIZE_TOPOLOGY||options.compute.OPTIMIZE_LENGTH || options.compute.OPTIMIZE_ROOT || options.compute.OPTIMIZE_COV1 
   || options.compute.OPTIMIZE_COV2 || options.compute.OPTIMIZE_COV3 || options.compute.OPTIMIZE_GAMMA || options.compute.OPTIMIZE_PI;

  bound=pow(10., -1.*(prec));
 /* for final estimation or for fixed topology (no-tree searching) */
  if (strcmp(status, "FINAL") == 0 || !options.compute.OPTIMIZE_TOPOLOGY) bound = 0.0001;


  /* INIT PARAMETERS */
  
  /* root */

  s_tree->froot=options.init.INIT_ROOT;    

  /* Gamma distribution */
  gamma_shp = s_tree->gamma_shp;
  gamma_clmean = check_alloc(gamma_nbcl, sizeof(double));
  init_gamma(gamma_shp, gamma_clmean);


  /* lengths */

  if(strcmp(options.init.INIT_LENGTH, "REDO")==0 || strcmp(status, "INITIAL")!=0)
    init_bl(s_tree);


  /* covarions */

  covar1=s_tree->cov1;
  covar2=s_tree->cov2; 
  covar3=s_tree->cov3;
  pi=s_tree->pi;


  /* random */
  
  if(options.init.NBRANDOM>0){    // no config file option for this
    if(options.init.NOISE<0){
      totallength=newtotallength=0.;
      for(i=0;i<2*nbseq-3;i++)
        totallength+=s_tree->listbr[i]->length;
      
      s_tree->froot=drand48();
      for(i=0;i<2*nbseq-3;i++){
        s_tree->listbr[i]->length=drand48();
        newtotallength+=s_tree->listbr[i]->length;
      }
      for(i=0;i<2*nbseq-3;i++){
        s_tree->listbr[i]->length*=(totallength/newtotallength);
      }
   }
    else{
      for(i=0;i<nbparam;i++){
	*(s_tree->var_param[i])*= 1+ (drand48()-0.5)*2.*options.init.NOISE/100.;
      }
    }
  }
  

  toadd=(double*)check_alloc(nbparam, sizeof(double));  
  paramvect=check_alloc(nbparam+1, sizeof(double));
  bestvect=(double*)check_alloc(nbparam+1, sizeof(double));
  finalvect=(double*)check_alloc(nbparam+1, sizeof(double));
  remvect=(double*)check_alloc(nbparam+1, sizeof(double));
  bounded=(int*)check_alloc(nbparam, sizeof(int));


  /* COMPUTE MAX LIKELIHOOD */

  reset_like(s_tree);
  nbiter=0; rem=0.;
  for(i=0;i<nbparam;i++) bestvect[i]=*(s_tree->var_param[i]);

  while(TRUE){ /* start of Newton-Raphson loop */
    nbiter++;
    if (verbose>=3) putchar('\n');
    if (verbose>=2) printf("nbiter=%d\told ln(L)=%f\n", nbiter, maxllike);
    firstcomp=1;
    nbreduction=0;
    for(i=0;i<nbparam;i++) remvect[i]=*(s_tree->var_param[i]);

    /* compute likelihood, compare to previous, */
    /* redo (up to 10 times) until it increases */
    while(firstcomp || llike<rem){

      for(i=0;i<nbparam;i++) {
	*(s_tree->var_param[i])=remvect[i]+toadd[i];
	bounded[i]=0;
      }
      
      boundaries(s_tree, bounded);   /* for deciding if doing simplex later */
      //      reset_tree(s_tree);    /* will do in minus_log_likelihood routine */

      for(i=0;i<nbparam;i++)
	paramvect[i]=*(s_tree->var_param[i]);

      llike=-minus_log_likelihood(paramvect);
      
      firstcomp=0;
      
      // if(nbiter<GAMMA_RANGE_ITER+2) break; /*in Galtier's NHML code*/
      if (nbiter<2) break;
      
      if(llike<rem){ 
        if (verbose>=2)
	  printf("likelihood decrease (%f->%f): moving backward\n", rem, llike);
	for(i=0;i<nbparam;i++) 
	  toadd[i]/=2.;
	
	nbreduction++;
	if(nbreduction==10) {
	  for(i=0;i<nbparam;i++) *(s_tree->var_param[i])=remvect[i]; 
	  /* if likelihood decrease 10 times no more modification of the parameters. 
	     restore the parameters to the values before the first time likelihood decreases. */
	  reset_tree(s_tree);
	  do_simplex_now=1; break;
	}
      }
    }
    
    if(do_simplex_now){
      do_simplex=1;
      break;
    }

    if(nbiter==1) maxllike=llike;
    else{
      if(maxllike<llike){
	maxllike=llike;
	for(i=0;i<nbparam;i++) bestvect[i]=paramvect[i];
      }
    }


    /* check positivity */

    if(llike>0.) {
      fputs("error: negative log\n", stderr);
      if (loglevel>=1){
        fputs("error: Negative log\n", out);
	//     output_tree_comp(root, out, NULL);
      }
      return 1.;
    }


    if (verbose>=3){
      printf("pi: %8.6f  #gamma: %d\tgamma_shp: %.6f\tlog likelihood: %.12g\n", pi, gamma_nbcl, gamma_shp, llike);
      printf("cov1: %.12f\tcov2: %.12f\tcov3: %.12f\n", covar1,covar2,covar3);
    }

    if(!OPTIMIZE) break;


    /* leave if converged */
    
    /* nbiter>10 allows always do more than 10 iterations*/
    /* perform simplex if some param is at the boundary of its definition interval */
    
    /* original code in procov1.0 */
    /*    if(nbiter!=1 && llike-rem<bound && llike-rem>-bound && nbiter>10){
      for(i=0;i<nbparam;i++)
	if(bounded[i]){ 
	  do_simplex=1;	
	}
      break;
    }
    */
    
    /* revised in procov2.0 in order to speed up tree search */
    /* this takes more time to converge */
    if (nbiter!=1) {
      if(strcmp(status,"FINAL")==0 || !options.compute.OPTIMIZE_TOPOLOGY) {
	if (llike-rem<bound && llike-rem>-(bound) && nbiter>10){ 
	  for(i=0;i<nbparam;i++)
	    if(bounded[i]){ 
	      do_simplex=1;	
	    }
	  break;
	}
      }
	
      else {  
	if (abs(llike-rem)<bound) {  /* this takes less time */
	  for(i=0;i<nbparam;i++)
	    if(bounded[i]){ 
	      do_simplex=1;
	    }
	break; 
	}
      }
    }


/* remember likelihood */
    rem=llike;
    s_tree->lkh=llike;


/* first steps: try several gamma distributions */
    old_gamma_shp=gamma_shp;
    if(options.compute.OPTIMIZE_GAMMA && nbiter<GAMMA_RANGE_ITER){
      best_gamma_l=llike; best_gamma=-1;
      for(i=0;i<GAMMA_RANGE_SIZE;i++){
        gamma_shp=gamma_range[i];
        DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);

	gamma_l=-minus_log_likelihood(paramvect);
	if(gamma_l>best_gamma_l){
          best_gamma=i;
          best_gamma_l=gamma_l;
        }
      }
      if(best_gamma!=-1){
        llike=best_gamma_l;
        rem=llike;
        s_tree->lkh=llike;
        gamma_shp=gamma_range[best_gamma];
      }
      else gamma_shp=old_gamma_shp;
      DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);

      minus_log_likelihood(paramvect);
    }
    

/* numerical derivatives for all parameters except pi */ 	
    derivatives(s_tree, paramvect);

/* not first steps: */
/* numerical derivative with respect to shape param of gamma distribution, */
/* and modification of gamma shape parameter. */
    if(options.compute.OPTIMIZE_GAMMA && nbiter>GAMMA_RANGE_ITER) {
    delta=0.00001;
    lll=s_tree->lkh;
    gamma_shp+=delta;
    DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);

    l1=-minus_log_likelihood(paramvect);
    gamma_shp-=2*delta;
    DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);

    l_1=-minus_log_likelihood(paramvect);
    gamma_shp+=delta;

    d_gamma=(l1-l_1)/(2*delta);
    d2_gamma=(l1+l_1-2*lll)/(delta*delta);  /* we changed Galtier's equation for 2nd deriv.*/

    den=1;
    old_gamma_shp=gamma_shp;
    if(d2_gamma<0.){ 
      toadd_gamma=-d_gamma/d2_gamma;
      while(den<1000){
        gamma_shp=old_gamma_shp+toadd_gamma/den;
 	if(gamma_shp<gammamin) gamma_shp=gammamin;
        DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
        lll=-minus_log_likelihood(paramvect);
        if(lll<llike) den*=2;   
        else break;
      }
    }
    if(den>1000 || d2_gamma>=0.){
      printf("den=%d\td2_gamma=%.10f\n",den,d2_gamma);
      gamma_shp=old_gamma_shp;
      DiscreteGamma (gamma_clmean, gamma_shp, gamma_shp, gamma_nbcl, 0);
    }
    else llike=lll;
    }


/* modify parameters */

    for(ii=0;ii<s_tree->nb_var_param;ii++){
      if (s_tree->deriv2[ii]==0.0) {
	if (s_tree->deriv[ii]==0.0) toadd[ii] = 0.0;
	else if (s_tree->deriv[ii]>0.0) 
	  switch(s_tree->param_nature[ii]){
	  case LENGTH: toadd[ii] = lmin-remvect[ii];break;
	  case COVA:case COVB:case COVC: toadd[ii] = covmin-remvect[ii];break;
	  case GaltierPI: toadd[ii] = pimin-remvect[ii]; break;
	  }
	
	else 
	  switch(s_tree->param_nature[ii]){
	  case LENGTH: toadd[ii] = lmax-remvect[ii];break;
	  case COVA:case COVB:case COVC: toadd[ii] = covmax-remvect[ii];break;
	  case GaltierPI: toadd[ii] = pimax-remvect[ii]; break;
	  }
      }
	
      else 
      	toadd[ii]=-s_tree->deriv[ii]/s_tree->deriv2[ii];
    }
    
    for(i=0;i<nbparam;i++){
      if(s_tree->deriv2[i]>0.){
	toadd[i]=0.;
	switch(s_tree->param_nature[i]){
	case ROOT: ctmp = "root"; break;
	case LENGTH: ctmp = "length"; break;
	case COVA: ctmp = "covar1"; break;
	case COVB: ctmp = "covar2"; break;
	case COVC: ctmp = "covar3"; break;
	case GaltierPI: ctmp = "pi"; break;
	default: ctmp = NULL;
	}
        if (verbose>=2) printf("param %d (%s) diverging\n", i+1, ctmp);
	if (loglevel>=3) fprintf(out, "iter:%d: param %d (%s) diverging\n", nbiter, i+1, ctmp);
      }
    }


  } /* end of Newton-Raphson loop */

  //  covarianceMatrix(s_tree, paramvect, gamma_shp, out);

  //  if(covar<=0.1) exact_cov=0;
  do_simplex = 0;    /* we disable doing simplex */           

	/* SIMPLEX */

  if (options.converge.SIMPLEX && do_simplex){
    if (verbose>=2)
      printf ("log likelihood for exact covarion before simplex: %.6f\n", maxllike);

    if (loglevel>=2){
      fprintf (out, "log likelihood for exact covarion before simplex: %.6f\n", maxllike);
      fflush(NULL);
    }

    if(verbose>=4) puts("Simplex...\nWarning: Simplex may take long time...\n");
    llike=simplex(bestvect, 0.01, finalvect);

    if(llike>maxllike){
      for(i=0;i<nbparam;i++)
        *(s_tree->var_param[i])=finalvect[i];
      reset_tree(s_tree);
      maxllike=llike;
    }
    else{
      for(i=0;i<nbparam;i++)
        *(s_tree->var_param[i])=bestvect[i];
      reset_tree(s_tree);
      llike=maxllike;
    }

    if (loglevel>=2)
      fprintf (out, "log likelihood for exact covarion after simplex: %.6f\n\n", llike);
  }

  //  getTreelength(&oriTreeLen, &treeLen);

  /* OUTPUT */
  
  s_tree->gamma_shp=gamma_shp;
  s_tree->cov1=covar1;
  s_tree->cov2=covar2;
  s_tree->cov3=covar3;
  s_tree->pi=pi;

  /*  write covarion site lkh and ras site lkh  */
  /* only for final tree or fixed tree */
  if (strcmp(status, "FINAL") == 0 || !options.compute.OPTIMIZE_TOPOLOGY) {
    if (siteLKHfileName!=NULL) {
      FILE *siteLKHfile;
      siteLKHfile=fopen(siteLKHfileName, "w");
 
      fprintf(siteLKHfile, "log(cov_plus_ras_sitelkh) ");
      for(i=0;i<s_tree->weightsum;i++) {
      j = s_tree->sitetopatt[i];
      fprintf(siteLKHfile, "%.6f ", log(s_tree->like_s[j]*pi+(1-pi)*s_tree->likeasrv_s[j]));
      }
      fputs("\n\n", siteLKHfile);

      fprintf(siteLKHfile, "log(cov_only_sitelkh) ");
      for(i=0;i<s_tree->weightsum;i++) {
	j = s_tree->sitetopatt[i];
	fprintf(siteLKHfile, "%.6f ", log(s_tree->like_s[j]));
      }
      fputs("\n\n", siteLKHfile);

      fprintf(siteLKHfile, "log(ras_only_sitelkh) ");
      for(i=0;i<s_tree->weightsum;i++) {
	j = s_tree->sitetopatt[i];
	fprintf(siteLKHfile, "%.6f ", log(s_tree->likeasrv_s[j]));
      }
      
      fputs("\n\n", siteLKHfile);
      fclose (siteLKHfile);
    }
  }
  
  ctree = (char*)check_alloc(50*nbseq, sizeof(char));
  if(ctree)
      stoc(s_tree->node, nbseq, 1, ctree);
  

	/* FREE */

  free(toadd);
  free(bestvect);
  free(paramvect);
  free(finalvect);
  free(remvect);
  free(bounded);
  free(gamma_clmean);

  return maxllike;
}
