/*  PROCOV (version 2.0): a program for maximum likelihood estimation of phylogenetic trees
 *  under the covarion models for protein sequences

 *  Copyright 2007 - 2009 by Huaichun Wang
 *
 *  Partially derived from NHML, a program for estimation of phylogenetic trees under non-homogenous 
 *  Markov models by N. Galtier and M. Gouy
 *
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include "mynhmlg.h"
#include <malloc.h>
#include <unistd.h>
#define MAXNTREE 100
#define MIN_NBLISTMAX 1000
#define MAX_NBLISTMAX 10000


char *infileName, *intreeName, *optfileName, *siteLKHfileName, *evaluated_fileName, *current_best_tree_fileName;
int SH_RESTART;

static const char *usage="usage: procov [options]\n"

"read a sequence and a tree; optimize branch lengths and calculate a likelihood\n"
"read the docs for more details\n\n"
"options: -h: this help\n"
"-V: version info\n"
"-l number: log file verbosity (default 3)\n"
"-v number: algorithm verbosity (default 1) (e.g. print spr number and likelihood for each spr)\n"
"-d number: debug/verbosity level (default 1)\n"
"-S: enable tree topology search (default no)\n"
"-m: protein substitution matrix (JTT, WAG or LG)\n"
"-i file: input file\n"
"-o file: output file\n"
"-t file: tree file\n"
"-c file: options config file\n"
"-L file: log likelihoods at sites for conbined COV+RAS, COV only and RAS only\n"
"-e file: specify a file that lists pre-evaluated trees, which will not be evaluated in tree search to save time\n"
"-b file: specify a file that writes currently best tree among the evaluated trees\n";

static const char *version="procov 2.0\n"
"by Huaichun Wang <hcwang@mathstat.dal.ca> and Peter Cordes <peter@cordes.ca>\n"
"Dalhousie University.  http://www.mathstat.dal.ca/Procov web site\n"
"based on NHML by Galtier & Gouy, 1998-2001\n";


int modelExists(char *model)
{
  static char *protModels[3] = {"JTT", "WAG", "LG"};
  int i;

  for(i = 0; i < 3; i++) {
    if (strcmp(model, protModels[i])==0) return i;
  }

  return -1;
}


void free_node(noeud nd, int nbcl)
{
  int j;

  free(nd->underlyingbr);
  free(nd->seq);
  if (nd->patternson1){		// leaf
    free(nd->patternson1);
    free(nd->patternson2);
    free(nd->x[0]); // pointers all point into a block allocated in one go
    free(nd->x);
  }
//  for(j=0;j<nd->nbpattern;j++) free(nd->x[j]);
//  for(j=0;j<nbcl;j++) free(nd->dx[j]);
//  free(nd->dx);
//  for(j=0;j<nbcl;j++) free(nd->d2x[j]);
//  free(nd->d2x);
//  for(j=0;j<nbcl;j++) free(nd->p1[j]);
  free(nd->p1);
//  for(j=0;j<nbcl;j++) free(nd->p2[j]);
  free(nd->p2);
//  for(j=0;j<nbcl;j++) free(nd->dp1[j]);
//  free(nd->dp1);
//  for(j=0;j<nbcl;j++) free(nd->dp2[j]);
//  free(nd->dp2);
//  for(j=0;j<nbcl;j++) free(nd->d2p1[j]);
//  free(nd->d2p1);
//  for(j=0;j<nbcl;j++) free(nd->d2p2[j]);
//  free(nd->d2p2);
  for(j=0;j<nbcl;j++) free(nd->rr1[j]);
  free(nd->rr1);
  for(j=0;j<nbcl;j++) free(nd->rr2[j]);
  free(nd->rr2);
  for(j=0;j<nbcl;j++) free(nd->drr1[j]);
  free(nd->drr1);
  for(j=0;j<nbcl;j++) free(nd->drr2[j]);
  free(nd->drr2);
  for(j=0;j<nbcl;j++) free(nd->d2rr1[j]);
  free(nd->d2rr1);
  for(j=0;j<nbcl;j++) free(nd->d2rr2[j]);
  free(nd->d2rr2);
  free(nd);
}

void free_tree(tree* s_tree, int nbcl){

  int i, nbseq, nbparam, lgseq;

  nbseq=s_tree->nbseq;
  lgseq=s_tree->lgseq;
  nbparam=s_tree->nb_var_param;

  free(s_tree->node[0]->x[0]);	// leaf nodes share a single x
  free(s_tree->node[0]->x);
  for(i=0;i<2*nbseq-2;i++)
    free_node(s_tree->node[i], nbcl);
  for(i=0;i<2*nbseq-3;i++)
    free(s_tree->listbr[i]);
  free(s_tree->listbr);
//  free(s_tree->listgc);
  free(s_tree->lgbr_init);
  free(s_tree->deriv2);
  free(s_tree->deriv);
  for(i=0;i<nbcl;i++) free(s_tree->classlike_s[i]);
  free(s_tree->classlike_s);
  free(s_tree->like_s);
  free(s_tree->likeasrv_s);
/* vestigial
  for(j=0;j<nbcl;j++){
    for(i=0;i<nbparam;i++){
      free(s_tree->classdlike_s[j][i]);
      free(s_tree->classd2like_s[j][i]);
    }
    free(s_tree->classdlike_s[j]);
    free(s_tree->classd2like_s[j]);
  }
  free(s_tree->classdlike_s);
  free(s_tree->classd2like_s);
*/
  for(i=0;i<nbparam;i++){
    free(s_tree->dlike_s[i]);
/* vestigial
    free(s_tree->dlikeasrv_s[i]);
    free(s_tree->d2like_s[i]);
    free(s_tree->d2likeasrv_s[i]);
*/
  }
  free(s_tree->alivebr);
  free(s_tree->dlike_s);
/* vestigial
  free(s_tree->dlikeasrv_s);
  free(s_tree->d2like_s);
  free(s_tree->d2likeasrv_s);
*/
  free(s_tree->weight);
  free(s_tree->var_param);
  free(s_tree->param_nature);
  for(i=0;i<nbseq;i++) free(s_tree->dist[i]);
  free(s_tree->dist);
  free(s_tree->sitetopatt);

}


void alloc_node(noeud nd, int lgseq, int nbparam, int nbseq, int leaf, int cross, int nbcl){

  int j, i;

  nd->underlyingbr=(int*)check_alloc(2*nbseq-3, sizeof(int));
  //  nd->underlyinggc=(int*)check_alloc(2*nbseq-2, sizeof(int));
  nd->seq=(char*)check_alloc(lgseq, sizeof(char));
  if(!leaf) nd->patternson1=(int*)check_alloc(lgseq, sizeof(int));
  if(!leaf) nd->patternson2=(int*)check_alloc(lgseq, sizeof(int));

  // x is allocated in setpattern(), when nbpattern is known.
  /*  nd->x=(like**)check_alloc(nbcl, sizeof(like*));
    nd->dx=(like**)check_alloc(nbcl, sizeof(like*));
    nd->d2x=(like**)check_alloc(nbcl, sizeof(like*));
    for(j=0;j<nbcl;j++) {
    nd->x[j]=(like*)check_alloc(lgseq, sizeof(like));
    nd->dx[j]=(like*)check_alloc(lgseq, sizeof(like));
    nd->d2x[j]=(like*)check_alloc(lgseq, sizeof(like));
    }   */

  
  if(!leaf){
    nd->p1 = check_alloc(nbcl*20 * nbcl*20, sizeof(*nd->p1));
    nd->p2 = check_alloc(nbcl*20 * nbcl*20, sizeof(*nd->p2));
  }

  /* nd->p1=(proba**)check_alloc(nbcl, sizeof(proba*));
  nd->p2=(proba**)check_alloc(nbcl, sizeof(proba*));
  nd->dp1=(dproba**)check_alloc(nbcl, sizeof(dproba*));
  nd->dp2=(dproba**)check_alloc(nbcl, sizeof(dproba*));
  nd->d2p1=(dproba**)check_alloc(nbcl, sizeof(dproba*));
  nd->d2p2=(dproba**)check_alloc(nbcl, sizeof(dproba*));
  
  for(j=0;j<nbcl;j++){
    nd->p1[j]=(proba*)check_alloc(nbcl, sizeof(proba));
    nd->p2[j]=(proba*)check_alloc(nbcl, sizeof(proba));
    nd->dp1[j]=(dproba*)check_alloc(nbcl, sizeof(dproba));
    nd->dp2[j]=(dproba*)check_alloc(nbcl, sizeof(dproba));
    nd->d2p1[j]=(dproba*)check_alloc(nbcl, sizeof(dproba));
    nd->d2p2[j]=(dproba*)check_alloc(nbcl, sizeof(dproba));
  }
  */

  nd->rr1=check_alloc(nbcl, sizeof(double*));
  nd->rr2=check_alloc(nbcl, sizeof(double*));
  nd->drr1=check_alloc(nbcl, sizeof(double*));
  nd->drr2=check_alloc(nbcl, sizeof(double*));
  nd->d2rr1=check_alloc(nbcl, sizeof(double*));
  nd->d2rr2=check_alloc(nbcl, sizeof(double*));

  for(i=0;i<nbcl;i++){
    nd->rr1[i]=check_alloc(nbcl, sizeof(double));
    nd->rr2[i]=check_alloc(nbcl, sizeof(double));
    nd->drr1[i]=check_alloc(nbcl, sizeof(double));
    nd->drr2[i]=check_alloc(nbcl, sizeof(double));
    nd->d2rr1[i]=check_alloc(nbcl, sizeof(double));
    nd->d2rr2[i]=check_alloc(nbcl, sizeof(double));
  }
}


static void err_mess(char* option){
  fprintf(stderr, "Missing or wrong line specifying option %s in option file\nDefault setting used\n", option);
}

/* refresh: remove all sites with gaps only from sequence alignment seq.
If option==0, all sites with at least one gap or with ambiguous amino acids
B, X, Z are removed. */
static void refresh(char** seq, int nbseq, int option){
  int lgseq, l=-1, drapeau, i, j, k;
  char **seqref ;

  lgseq=(int)strlen(seq[0]);
  seqref=(char**)check_alloc(nbseq, sizeof(char*));
  for(i=0;i<nbseq;i++)
     seqref[i]=(char*)check_alloc(lgseq+1, sizeof(char));

  if (option==0){
    for(i=0;i<lgseq;i++){
      drapeau=0;
      for(j=0;j<nbseq;j++){
	// if (strchr("ABCDEFGHIKLMNPQRSTVWXYZ", seq[j][i]) == NULL) {
        if (strchr("ACDEFGHIKLMNPQRSTVWY", seq[j][i]) == NULL) {
	  drapeau=1;
          break;
        }
      }
      if (drapeau==0){
	l++;
	for(k=0;k<nbseq;k++){
    	  *(seqref[k]+l)=*(seq[k]+i);
	}
      }
    }
  }
  else{
    for(i=0;i<lgseq;i++){
      drapeau=0;
      for(j=0;j<nbseq;j++){
	//	if (strchr("ABCDEFGHIKLMNPQRSTVWXYZ", *(seq[j]+i))) {
	if (strchr("ACDEFGHIKLMNPQRSTVWY", *(seq[j]+i))) {
	  drapeau=1;
	  break;
	}
      }
      if (drapeau==1){
	l++;
	for(k=0;k<nbseq;k++)
	  if (strchr("ARNDCQEGHILKMFPSTWYVBZ",seq[k][i]))
	    *(seqref[k]+l)=*(seq[k]+i);
	  else
	    *(seqref[k]+l)='?';
      }
    }
  }

  for(i=0;i<nbseq;i++)
    for (j=l+1;j<lgseq;j++)
      *(seqref[i]+j)='\0';
  for (i=0;i<nbseq;i++)
    for (j=0;j<lgseq;j++)
      *(seq[i]+j)=*(seqref[i]+j);
  for(i=0;i<nbseq;i++) free(seqref[i]);
  free(seqref);
}


void set_var_param(tree* s_tree)
{
  int i, j, k;

  s_tree->nb_var_param=0;
  if(options.compute.OPTIMIZE_ROOT) s_tree->nb_var_param++;
  if(options.compute.OPTIMIZE_LENGTH) s_tree->nb_var_param+=s_tree->nbalivebr;
  if(options.compute.OPTIMIZE_COV1) s_tree->nb_var_param++;
  if(options.compute.OPTIMIZE_COV2) s_tree->nb_var_param++;
  if(options.compute.OPTIMIZE_COV3) s_tree->nb_var_param++;
  if(options.compute.OPTIMIZE_PI) s_tree->nb_var_param++;

  s_tree->var_param=(double**)check_alloc(s_tree->nb_var_param, sizeof(double*));
  s_tree->param_nature=(int*)check_alloc(s_tree->nb_var_param, sizeof(int));
  for(i=0;i<nb_param_type;i++) s_tree->which_var_param[i]=-1;   // nb_param_type defined in mynhml.h
  i=0;

  if(options.compute.OPTIMIZE_ROOT){
    s_tree->var_param[i]=&(s_tree->froot);
    s_tree->param_nature[i]=ROOT;
    s_tree->which_var_param[ROOT]=i;
    i++;
  }

  if(options.compute.OPTIMIZE_LENGTH){
    k=0;
    for(j=0;j<2*s_tree->nbseq-3;j++){
      if(s_tree->alivebr[j]){
	s_tree->var_param[i+k]=&(s_tree->listbr[j]->length);
	s_tree->param_nature[i+k]=LENGTH;
	k++;
      }
    }
    s_tree->which_var_param[LENGTH]=i;
    i+=s_tree->nbalivebr;
  }

 /* if we change s_tree->var_param[i] covar1, 2, 3 and pi will be changed; they are same;
covar1, 2, 3 and pi are used in likelihood calculation in compute() */
  if(options.compute.OPTIMIZE_COV1){
    s_tree->var_param[i]=&covar1;
    s_tree->param_nature[i]=COVA;
    s_tree->which_var_param[COVA]=i;
    i++;
  }
  if(options.compute.OPTIMIZE_COV2){
    s_tree->var_param[i]=&covar2;
    s_tree->param_nature[i]=COVB;
    s_tree->which_var_param[COVB]=i;
    i++;
  }
  if(options.compute.OPTIMIZE_COV3){
    s_tree->var_param[i]=&covar3;
    s_tree->param_nature[i]=COVC;
    s_tree->which_var_param[COVC]=i;
    i++;
  }

  if(options.compute.OPTIMIZE_PI){
    s_tree->var_param[i]=&pi;  
    s_tree->param_nature[i]=GaltierPI;
    s_tree->which_var_param[GaltierPI]=i;
  }
}


void set_var_param1(tree* s_tree) {
  int i,j,k;

  if(options.compute.OPTIMIZE_LENGTH) s_tree->nb_var_param+=s_tree->nbalivebr;

  s_tree->var_param=(double**)check_alloc(s_tree->nb_var_param, sizeof(double*));
  s_tree->param_nature=(int*)check_alloc(s_tree->nb_var_param, sizeof(int));
  for(i=0;i<nb_param_type;i++) s_tree->which_var_param[i]=-1;   // nb_param_type defined in mynhml.h
  
  i=0;
  if(options.compute.OPTIMIZE_LENGTH){
    k=0;
    for(j=0;j<2*s_tree->nbseq-3;j++){
      if(s_tree->alivebr[j]){
	s_tree->var_param[i+k]=&(s_tree->listbr[j]->length);
	s_tree->param_nature[i+k]=LENGTH;
	k++;
      }
    }
    s_tree->which_var_param[LENGTH]=i;
    i+=s_tree->nbalivebr;
  }

  options.compute.OPTIMIZE_GAMMA = FALSE;
  options.compute.OPTIMIZE_COV1 = FALSE;
  options.compute.OPTIMIZE_COV2 = FALSE;
  options.compute.OPTIMIZE_COV3 = FALSE;
  options.compute.OPTIMIZE_PI = FALSE;
}


static void getoptions(char* filename)
{
  char *text, *prov;
  int i, a, ret, model;
  FILE *in_opt;
  
  procov_options *opt = &options;


/* default options */
  opt->compute.OPTIMIZE_LENGTH=TRUE;
  opt->compute.OPTIMIZE_ROOT=FALSE;
  opt->compute.OPTIMIZE_GAMMA=TRUE;
  opt->compute.OPTIMIZE_COV1=FALSE;
  opt->compute.OPTIMIZE_COV2=FALSE;
  opt->compute.OPTIMIZE_COV3=FALSE;
  opt->compute.OPTIMIZE_PI=FALSE;
  strcpy(opt->init.INIT_LENGTH, "REDO");
  opt->init.INIT_ROOT=-1.;
  opt->init.NBRANDOM=0;
  opt->init.NOISE=0;
  opt->init.INIT_GAMMA=-1.;
  opt->init.GAMMA_NCL=4;
  opt->init.INIT_COV1=-1.;
  opt->init.INIT_COV2=-1.;
  opt->init.INIT_COV3=-1.;
  opt->init.INIT_PI=-1.;
  opt->converge.PRECISION=4;
  opt->converge.SIMPLEX=TRUE;
  opt->print.EVAL_OUT=FALSE;
  opt->print.OUTPUT_COMP=FALSE;
  opt->print.COV_SITE_PATTERN=FALSE;
  opt->SH_G=-1;
  opt->SH_MAXLCROSSED=lmax;
  opt->SH_MAXBOOTCROSSED=INT_MAX;

  text=(char*)check_alloc(8000, sizeof(char));

  in_opt=fopen(filename, "r");
  if(!in_opt) {
    fprintf(stderr, "problem opening options file %s", filename);
    perror(" "); exit(1);
  }

  i=fread(text, sizeof(char), 7999, in_opt);
  if(ferror(in_opt)){
    perror("error reading options file");  exit(1);
  }
  text[i]='\0';

  fclose(in_opt);


  prov=strstr(text, "INIT_LENGTH");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_LENGTH"); goto fininitlg; }
  if(prov)
    prov++;
  else { err_mess("INIT_LENGTH"); goto fininitlg; }
  while(*prov==' ') prov++;
  i=0;
  while(*prov!=' ' && *prov!='\t' && *prov!='\n'){
    opt->init.INIT_LENGTH[i]=*prov;
    i++; prov++;
  }
  opt->init.INIT_LENGTH[i]=0;
  if(strcmp(opt->init.INIT_LENGTH, "KEEP") && strcmp(opt->init.INIT_LENGTH, "REDO")){
    err_mess("INIT_LENGTH"); strcpy(opt->init.INIT_LENGTH, "REDO");
  }
  fininitlg:

  prov=strstr(text, "OPTIMIZE_LENGTH");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_LENGTH"); goto finol; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_LENGTH"); goto finol; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_LENGTH));
  if(opt->compute.OPTIMIZE_LENGTH!=0 && opt->compute.OPTIMIZE_LENGTH!=1) {
    err_mess("OPTIMIZE_LENGTH"); opt->compute.OPTIMIZE_LENGTH=1;
  }
  finol:

  
  prov=strstr(text, "OPTIMIZE_ROOT");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_ROOT"); goto finor; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_ROOT"); goto finor; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_ROOT));
  if(opt->compute.OPTIMIZE_ROOT!=0 && opt->compute.OPTIMIZE_ROOT!=1) {
    err_mess("OPTIMIZE_ROOT"); opt->compute.OPTIMIZE_ROOT=1;
  }
  finor:


  prov=strstr(text, "OPTIMIZE_GAMMA");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_GAMMA"); goto finog; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_GAMMA"); goto finog; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_GAMMA));
  if(opt->compute.OPTIMIZE_GAMMA!=0 && opt->compute.OPTIMIZE_GAMMA!=1) {
    err_mess("OPTIMIZE_GAMMA"); opt->compute.OPTIMIZE_GAMMA=1;
  }
  finog:


  prov=strstr(text, "OPTIMIZE_COV1");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_COV_off/on"); goto finocov1; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_COV_off/on"); goto finocov1; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_COV1));
  if(opt->compute.OPTIMIZE_COV1!=0 && opt->compute.OPTIMIZE_COV1!=1) {
    err_mess("OPTIMIZE_COV_off/on"); opt->compute.OPTIMIZE_COV1=0;
  }
  finocov1:

  prov=strstr(text, "OPTIMIZE_COV2");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_COV_on/off"); goto finocov2; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_COV_on/off"); goto finocov2; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_COV2));
  if(opt->compute.OPTIMIZE_COV2!=0 && opt->compute.OPTIMIZE_COV2!=1) {
    err_mess("OPTIMIZE_COV_on/off"); opt->compute.OPTIMIZE_COV2=0;
  }
  finocov2:

  prov=strstr(text, "OPTIMIZE_COV3");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_COV_on/on"); goto finocov3; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_COV_on/on"); goto finocov3; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_COV3));
  if(opt->compute.OPTIMIZE_COV3!=0 && opt->compute.OPTIMIZE_COV3!=1) {
    err_mess("OPTIMIZE_COV_on/on"); opt->compute.OPTIMIZE_COV3=0;
  }
  finocov3:

  prov=strstr(text, "OPTIMIZE_PI");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("OPTIMIZE_PI"); goto finopi; }
  if(prov)
    prov++;
  else { err_mess("OPTIMIZE_PI"); goto finopi; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->compute.OPTIMIZE_PI));
  if(opt->compute.OPTIMIZE_PI!=0 && opt->compute.OPTIMIZE_PI!=1) {
    err_mess("OPTIMIZE_PI"); opt->compute.OPTIMIZE_PI=0;
  }
  finopi:

  /*
  prov=strstr(text, "INIT_ROOT");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_ROOT"); goto finkrl; }
  if(prov)
    prov++;
  else { err_mess("INIT_ROOT"); goto finkrl; }
  while(*prov==' ') prov++;
  a=sscanf(prov, "%le", &(opt->init.INIT_ROOT));
  if( opt->init.INIT_ROOT>1. || a<1){
    err_mess("INIT_ROOT"); opt->init.INIT_ROOT=-1.;
  }
  finkrl:
  */

  prov=strstr(text, "INIT_COV1");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_COV1"); goto finico1; }
  if(prov)
    prov++;
  else { err_mess("INIT_COV1"); goto finico1; }
  while(*prov==' ') prov++;
  a=sscanf(prov, "%le", &(opt->init.INIT_COV1));
  if(a<1){
    err_mess("INIT_COV1"); opt->init.INIT_COV1=-1.;
  }
  finico1:

  prov=strstr(text, "INIT_COV2");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_COV2"); goto finico2; }
  if(prov)
    prov++;
  else { err_mess("INIT_COV2"); goto finico2; }
  while(*prov==' ') prov++;
  a=sscanf(prov, "%le", &(opt->init.INIT_COV2));
  if(a<1){
    err_mess("INIT_COV2"); opt->init.INIT_COV2=-1.;
  }
  finico2:

  prov=strstr(text, "INIT_COV3");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_COV3"); goto finico3; }
  if(prov)
    prov++;
  else { err_mess("INIT_COV3"); goto finico3; }
  while(*prov==' ') prov++;
  a=sscanf(prov, "%le", &(opt->init.INIT_COV3));
  if(a<1){
    err_mess("INIT_COV3"); opt->init.INIT_COV3=-1.;
  }
  finico3:


  prov=strstr(text, "INIT_PI");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_PI"); goto finipi; }
  if(prov)
    prov++;
  else { err_mess("INIT_PI"); goto finipi; }
  while(*prov==' ') prov++;
  a=sscanf(prov, "%le", &(opt->init.INIT_PI));
  if(a<1){
    err_mess("INIT_PI"); opt->init.INIT_PI=-1.;
  }
  finipi:


  prov=strstr(text, "PRECISION");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("PRECISION"); goto finpr; }
  if(prov)
    prov++;
  else { err_mess("PRECISION"); goto finpr; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->converge.PRECISION));
  if(opt->converge.PRECISION<=-5) {
    err_mess("PRECISION"); opt->converge.PRECISION=4;
  }
  finpr:


  prov=strstr(text, "SIMPLEX");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("SIMPLEX"); goto finsp; }
  if(prov)
    prov++;
  else { err_mess("SIMPLEX"); goto finsp; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->converge.SIMPLEX));
  if(opt->converge.SIMPLEX!=0 && opt->converge.SIMPLEX!=1) {
    err_mess("SIMPLEX"); opt->converge.SIMPLEX=1;
  }
  finsp:


  prov=strstr(text, "EVAL_OUT");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("EVAL_OUT"); goto finevo; }
  if(prov)
    prov++;
  else { err_mess("EVAL_OUT"); goto finevo; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->print.EVAL_OUT));
  if(opt->print.EVAL_OUT!=0 && opt->print.EVAL_OUT!=1) {
    err_mess("EVAL_OUT"); opt->print.EVAL_OUT=1;
  }

  finevo:


  prov=strstr(text, "COV_SITE_PATTERN");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("COV_SITE_PATTERN"); goto fincsp; }
  if(prov)
    prov++;
  else { err_mess("COV_SITE_PATTERN"); goto fincsp; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->print.COV_SITE_PATTERN));
  if(opt->print.COV_SITE_PATTERN!=0 && opt->print.COV_SITE_PATTERN!=1) {
    err_mess("COV_SITE_PATTERN"); opt->print.COV_SITE_PATTERN=1;
  }

  fincsp:



  prov=strstr(text, "SH_G");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("SH_G"); goto finshg; }
  if(prov)
    prov++;
  else { err_mess("SH_G"); goto finshg; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->SH_G));

  finshg:

  prov=strstr(text, "SH_MAXLCROSSED");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("SH_MAXLCROSSED"); goto finshm; }
  if(prov)
    prov++;
  else { err_mess("SH_MAXLCROSSED"); goto finshm; }
  while(*prov==' ') prov++;
  sscanf(prov, "%le", &(opt->SH_MAXLCROSSED));
  if(opt->SH_MAXLCROSSED<=0) opt->SH_MAXLCROSSED=lmax;

 finshm:


  prov=strstr(text, "SH_MAXBOOTCROSSED");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("SH_MAXBOOTCROSSED"); goto finshb; }
  if(prov)
    prov++;
  else { err_mess("SH_MAXBOOTCROSSED"); goto finshb; }
  while(*prov==' ') prov++;
  sscanf(prov, "%d", &(opt->SH_MAXBOOTCROSSED));

  if(opt->SH_MAXBOOTCROSSED<=0) opt->SH_MAXBOOTCROSSED=INT_MAX;

  finshb:



  prov=strstr(text, "INIT_GAMMA");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("INIT_GAMMA"); goto fingam; }
  if(prov)
    prov++;
  else { err_mess("INIT_GAMMA"); goto fingam; }
  while(*prov==' ') prov++;
  ret=sscanf(prov, "%le", &(opt->init.INIT_GAMMA));
  if(ret==0) { err_mess("INIT_GAMMA"); opt->init.INIT_GAMMA=-1.;}

  fingam:


  prov=strstr(text, "GAMMA_NBCLASS");
  if(prov)
    prov=strchr(prov, '=');
  else { err_mess("GAMMA_NBCLASS"); goto fingac; }
  if(prov)
    prov++;
  else { err_mess("GAMMA_NBCLASS"); goto fingac; }
  while(*prov==' ') prov++;
  ret=sscanf(prov, "%d", &(opt->init.GAMMA_NCL));
  if(ret==0 || opt->init.GAMMA_NCL<1){
    fputs("GAMMA_NBCLASS less than 1 or not defined - using as default 4 gamma rate categories", stderr);
    opt->init.GAMMA_NCL=4;
  }
  fingac:


  if(opt->init.INIT_GAMMA<0. && opt->compute.OPTIMIZE_GAMMA==0)
    opt->init.GAMMA_NCL=1;

  free(text);
}




static int samesite(char* site, tree* s_tree, int i){
  int j;
  for(j=0;j<s_tree->nbseq;j++)
    if(s_tree->node[j]->seq[i]!=site[j]) return 0;
  return 1;
}



int main(int argc, char** argv)
{
  int i, j, nbseq, nbtree, maxtreesize, nbeval;
  int secondes, minutes, heures, jours;
  char **seq, **seqname;
  char *prov, *prov2, protModel[20];
  double runtime;
  char *resultFileName=NULL;
  char *ctree, *current_best, **evaluated, *ret;
  int model, modelFlag, flag=0;
  FILE *treefile,*in, *optfile, *current_best_tree_file, *evaluated_file;
  time_t debut, fin;
  do_treesearch = 0;


/* printmemory_init(); */
#ifdef M_TOP_PAD
  mallopt(M_TOP_PAD, 10<<20); // get memory in 10MB chunks
  mallopt(M_TRIM_THRESHOLD, 4<<20);
  mallopt(M_MMAP_MAX, 0);	// never mmap.  http://math-atlas.sourceforge.net/errata.html#malloc
#endif

  do_tree_maxjobs = 1; // defaults
  debug = 1;
  verbose = 1;
  loglevel = 3;
  SH_RESTART=0;
  
  if (argc<2) {
    printf ("You need to specify a protein substitution model, an input sequence file, an initial tree file and an option config file\n\nType procov -h for help\n");
    exit(0);
  }

  
  // Parse command line options
  opterr = 1; // make getopt print specific error messages for us
  while ((i = getopt (argc, argv, "hVSX:l:v:d:D:m:i:o:t:c:L:e:b:")) != -1){
    switch(i){
    case 'h':
      puts(usage);
      return 0;
    case 'V':
      puts(version);
      return 0;
    case 'v': verbose=atoi(optarg); break;
    case 'l': loglevel=atoi(optarg); break;
    case 'd': debug=atoi(optarg); break;
    case 'S': do_treesearch = 1;
	    loglevel = max(loglevel, 1); break;
    case 'X': if (strchr(optarg, 'L')) do_tree_likeonly = 1; break;
    case 'm': 
      strcpy(protModel, optarg);
      model = modelExists(protModel);
      if(model == -1) {
	printf("Model %s does not exist\n\n", protModel);
	printf("Available models: JTT, WAG and LG\n");
	exit(0);
      }
      break;
    case 'i': infileName = optarg; break;
    case 'o': resultFileName = optarg; break;
    case 't': intreeName = optarg; break;
    case 'c': optfileName = optarg; break;
    case 'L': siteLKHfileName = optarg; break;
    case 'e': 
      evaluated_fileName = optarg;
      if (evaluated_fileName[0] == '-') {fputs("you need an evaluated_file name after -e\n", stderr); exit(0);}
      SH_RESTART=1; break;
    case 'b': current_best_tree_fileName = optarg; 
      if (current_best_tree_fileName[0] == '-') {fputs("you need a current_best_tree file name after -b\n", stderr); exit(0);}
      flag = 1;
      break;
    case '?':
      fputs("you need -h (help)\n", stderr);
      return 1;
    default:
      abort ();	// this should never happen
    }
  }

  modelFlag=0;
  for (i=0;i<argc;i++)
    if (strstr(argv[i], "-m")!=NULL) {
      modelFlag=1;
      break;}
  
  if (modelFlag==0) {
    printf ("You need to specify a protein substitution model\nType procov -h for help\n"); exit(0);
  }

  srand48(seed);


  /** input sequences. exit() on file error **/
  nbseq=readSequencefile(infileName, &seq, &seqname, TRUE);

  refresh(seq, nbseq, 1);


   /** input trees **/
  treefile=fopen(intreeName, "r");
  if(!treefile) {
    fprintf(stderr, "cannot open tree file: %s\nType procov -h for help\n", intreeName);
    /*   perror(" ");*/ exit(1);
  }
  ctree = readtree(treefile, nbseq);


   /** input options **/
  getoptions(optfileName);	// set globals options struct
  if (do_treesearch) options.compute.OPTIMIZE_TOPOLOGY=TRUE;
  else options.compute.OPTIMIZE_TOPOLOGY=FALSE;
  if (!strchr(ctree, ':')) strcpy(options.init.INIT_LENGTH, "REDO");  // if the input tree without branch lengths, redo the lengths

  /** output file **/
  if (!resultFileName)
    asprintf(&resultFileName, "%s.result.dat", infileName);
  out=fopen(resultFileName, "w");
  if(!out){
    fputs("procov: problem creating output file:", stderr);
    perror(resultFileName); exit(1);
  }


  if (do_treesearch == 0) 
    SH_RESTART = 0;
  else {
  /** input saved work **/

  if (SH_RESTART) {
    evaluated_file = fopen(evaluated_fileName, "r");
    
    if(!evaluated_file){
      printf("\nCannot find evaluated file %s\n", evaluated_fileName);
      printf("but I am going to write a list of the evaluated trees to %s\n", evaluated_fileName);
      evaluated=NULL; nbeval=0;
    }
      
    else {
      flag = 2;  /* if there is an evaluated file a current_best_tree file must exist */
      evaluated=check_alloc(MAX_NBLISTMAX, sizeof(char*));
      nbeval=0; j=3*MAXLNAME*nbseq;
      while(1){
	evaluated[nbeval]=check_alloc(j, sizeof(char));
	ret=fgets(evaluated[nbeval], j, evaluated_file);
	if(ret==NULL) break;
	nbeval++;
      }
    }
  }

  else {evaluated=NULL; nbeval=0;}
  

  if (SH_RESTART) {
    if (flag==0) {
      printf("\nthe -b for the current_best_tree_file is not specified\n");
      printf("but I am going to write it to a file called current_best_tree\n");
      current_best_tree_fileName = "current_best_tree";
    }
 
    else {
      current_best_tree_file = fopen(current_best_tree_fileName, "r");
      if (!current_best_tree_file) {
	if (flag == 2) {
	  printf("cannot open the current_best_tree file: %s\n", current_best_tree_fileName);
	  exit(0);
	}
	else {
	  printf("Cannot find the current_best_tree file\n");
	  printf("but I am going to write the current_best_tree to %s\n", current_best_tree_fileName);
	}
      }
      
      else ctree = readtree(current_best_tree_file, nbseq);
    }
  }
  }


  time(&debut);


  fprintf(out,"PROCOV report\n\n");

  for (i=0;i<argc;i++)
    fprintf(out,"%s ", argv[i]);

  fprintf(out,"\n\nprotein model: %s\n", protModel);


  precomputeProtMat(model);


  /** shake: do tree search and evaluation **/

  shake(nbseq, seq, seqname, ctree, evaluated, nbeval, out);

  printf("Detailed results are written into file: %s\n", resultFileName);

  time(&fin);

  runtime=difftime(fin, debut);

    jours=(int)(runtime/86400.);
    heures=(int)((runtime-86400*jours)/3600.);
    minutes=(int)((runtime-86400*jours-3600*heures)/60.);
    secondes=(int)(runtime-86400*jours-3600*heures-60*minutes);
    printf("(running time : ");
    if(jours) printf("%d days, ", jours);
    if(jours || heures) printf("%d hours, ", heures);
    if(jours || heures || minutes) printf("%d minutes, ", minutes);
    printf("%d seconds)\n\n", secondes);

    fprintf(out, "(running time : ");
    if(jours) fprintf(out, "%d days, ", jours);
    if(jours || heures) fprintf(out, "%d hours, ", heures);
    if(jours || heures || minutes) fprintf(out, "%d minutes, ", minutes);
    fprintf(out, "%d seconds)\n\n", secondes);

  fclose(out);
  
}
